from typing import Dict, List

import numpy as np
from matplotlib import pyplot as plt

from qat.purr.backends.qblox.codegen import QbloxPackage
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


def plot_packages(packages: List[QbloxPackage]):
    if not packages:
        return

    max_length = max([len(pkg.timeline) for pkg in packages])
    if max_length <= 0:
        return

    # Padding short timelines with zeros
    for pkg in packages:
        length = len(pkg.timeline)
        if length < max_length:
            pkg.timeline = np.append(
                pkg.timeline, np.zeros(max_length - length, dtype=pkg.timeline.dtype)
            )

    t = np.linspace(0, max_length, max_length)
    fig, axes = plt.subplots(
        nrows=len(packages),
        ncols=1,
        sharex=False,
        sharey=False,
        squeeze=False,
        figsize=(10, 5),
    )
    fig.suptitle("Target timeline plots")
    for i, pkg in enumerate(packages):
        axes[i][0].plot(t, pkg.timeline.real, label="I")
        axes[i][0].plot(t, pkg.timeline.imag, label="Q")
        axes[i][0].set_title(pkg.target)
        axes[i][0].set_xlabel("Time (ns)")
        axes[i][0].set_ylabel("Digital offset")
        axes[i][0].autoscale()
        axes[i][0].legend()

    plt.tight_layout()
    plt.show()


def plot_acquisitions(acquisitions: Dict, *args, **kwargs):
    if not acquisitions:

        return

    integration_length: int = kwargs["integration_length"]

    for i, (acq_name, acq) in enumerate(acquisitions.items()):
        fig, axes = plt.subplots(
            nrows=2,
            ncols=1,
            sharex=False,
            sharey=False,
            squeeze=False,
            figsize=(10, 5),
        )
        fig.suptitle(f"Scope and Acquisition plots for {acq_name}")

        # Scope acquisition
        scope_acq_i = np.array(acq["acquisition"]["scope"]["path0"]["data"])
        scope_acq_q = np.array(acq["acquisition"]["scope"]["path1"]["data"])
        max_length = scope_acq_i.size
        if max_length > 0:
            t = np.linspace(0, max_length, max_length)
            axes[0, 0].plot(t, scope_acq_i, label="I")
            axes[0, 0].plot(t, scope_acq_q, label="Q")
            axes[0, 0].set_xlabel("Sample (ns)")
            axes[0, 0].set_ylabel("Scope signal")
            axes[0, 0].autoscale()
            axes[0, 0].legend()

        # Binned acquisition
        bin_acq_i = (
            np.array(acq["acquisition"]["bins"]["integration"]["path0"])
            / integration_length
        )
        bin_acq_q = (
            np.array(acq["acquisition"]["bins"]["integration"]["path1"])
            / integration_length
        )
        max_length = bin_acq_i.size
        if max_length > 0 and not np.isnan(np.append(bin_acq_i, bin_acq_q)).any():
            t = np.linspace(0, max_length, max_length)
            axes[1, 0].plot(t, bin_acq_i, label="I")
            axes[1, 0].plot(t, bin_acq_q, label="Q")
            axes[1, 0].set_xlabel("Shot")
            axes[1, 0].set_ylabel("Acq Signal")
            axes[1, 0].autoscale()
            axes[1, 0].legend()

        plt.tight_layout()
        plt.show()



from abc import ABC, abstractmethod
from typing import List

from qat.ir.result_base import ResultManager
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class PassConcept(ABC):
    """
    Base class describing the abstraction of a pass.

    See PassManager.
    """

    @abstractmethod
    def run(self, ir, res_mgr: ResultManager, *args, **kwargs):
        pass


class PassModel(PassConcept):
    """
    Implements the polymorphic pass API. A wrapper for any object providing a run()
    method that accepts some unit of IR as well as a ResultManager.

    See PassManager.
    """

    def __init__(self, pass_obj):
        self._pass = pass_obj

    def run(self, ir, res_mgr: ResultManager, *args, **kwargs):
        return self._pass.run(ir, res_mgr, *args, **kwargs)


class PassInfoMixin(ABC):
    """
    Base mixin specifying pass identification mechanism. A pass has an id and name.

    Such identification is loosely specified as it related closely to how passes and their results
    are managed throughout their lifecycle.

    See PassManager.
    """

    def id(self):
        pass

    def name(self):
        pass


class AnalysisPass(PassInfoMixin):
    """
    Base class of all passes that compute some form of analysis on the IR.
    The IR is imperatively left intact.
    """

    def run(self, ir, res_mgr: ResultManager, *args, **kwargs):
        pass


class TransformPass(PassInfoMixin):
    """
    Base class for all passes that mutate the IR in-place.
    """

    def run(self, ir, res_mgr: ResultManager, *args, **kwargs):
        pass


class ValidationPass(PassInfoMixin):
    """
    Base class for all passes that verify or run some form of check (i.e semantics, legalisation, ...)
    on the IR.

    Behaviour is loose as to what the result of the validation would be: either a result or an error raised.
    It can change according to circumstances.
    """

    def run(self, ir, res_mgr: ResultManager, *args, **kwargs):
        pass


class PassManager(PassInfoMixin):
    """
    Represents a sequential composite of passes, which can be composites themselves. Although not explicitly
    specified, the pass manager is best modelled as a DAG that acts as a pass itself. In doing so, it runs
    a sequence of (composite or leaf) passes over some unit of IR and aggregates results from them.

    Result aggregation (in general) can be achieved via side effects or simply return values. We adopted
    the latter approach where a ResultManager is passed in as argument. The rationale behind this choice
    is two-fold:

    1) Allow passes to peek, use, or potentially invalidate results via the ResultManager without burdening
    their parent PassManager.

    2) Let the passes and results define their dependency graphs on the fly which saves us from explicitly
    building and maintaining them. Here it's also a possible direction to let passes lazily declare, create,
    and run their dependencies as they get discovered.

    Result aggregation would also potentially involve cache handling as described in the ResultManager.
    This calls for proper pass identification, cache invalidation, and pass/result cycle detection techniques.

    Example of a DAG:

    (PM1 A B (PM2 C D ) E F) can be thought of as:

            <---PM2--->

               --> C
    A --> B --|    ↓
               --> D --> E --> F

    <-----------PM1------------>

    Although the notation of a given (sub-) DAG looks sequential, passes within can run sequentially
    or in parallel depending on their inter-dependencies.

    Notes thus far describe an ideal fully-fledged PassManager. However, today's needs are very simple as we barely
    have visible features of modular quantum compilation workflow. It is therefore wise to keep the implementation
    light and basic where component passes run and register their own results within the ResultManager passed
    in as argument. Another reason is the possibility in the near future to adopt and tap into existing and mature
    pass pipeline infrastructure which saves us from reinventing or rediscovering the same concepts already covered
    by the opensource community.
    """

    def __init__(self):
        self.passes: List[PassModel] = []

    def run(self, ir, res_mgr: ResultManager, *args, **kwargs):
        for p in self.passes:
            p.run(ir, res_mgr, *args, **kwargs)

    def add(self, pass_obj):
        self.passes.append(PassModel(pass_obj))
        return self

    def __or__(self, pass_obj):
        return self.add(pass_obj)


class InvokerMixin(ABC):
    """
    Useful for compilation (global design-wise) stages/phases that are meant to invoke some
    arbitrary formation of passes in the form of a pipeline.

    Organisation, registering, visibility, and discovery of pipelines w.r.t the global quantum compilation
    workflow is hard to pin down early on in this design partly because today's needs are trivial and
    the traditional "hand-me-down" cascade-like compilation stages is likely to shift as QAT matures.

    However, we'll start by specifying that an invoker builds and runs its own pipeline, validates its
    analyses results, and uses those results for its purpose.
    """

    @abstractmethod
    def build_pass_pipeline(self, *args, **kwargs) -> PassManager:
        pass

    def run_pass_pipeline(self, ir, res_mgr: ResultManager, *args, **kwargs):
        return self.build_pass_pipeline(*args, **kwargs).run(ir, res_mgr, *args, **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
from __future__ import annotations

import os
import sys
from enum import Enum, auto
from typing import Dict, List, Optional, Set, TypeVar, Union

import jsonpickle
import jsonpickle.ext.numpy as jsonpickle_numpy
import numpy as np
from jsonpickle import Pickler, Unpickler
from jsonpickle.util import is_picklable

from qat.purr.utils.logger import get_default_logger

# TODO: Remove the setting of the recursion limit as soon as the pickling of hardware is
# not blocked by the default limit.
sys.setrecursionlimit(2000)

log = get_default_logger()
jsonpickle_numpy.register_handlers()


def build_resonator(
    resonator_id,
    resonator_channel,
    *args,
    measure_fixed_if=False,
    acquire_fixed_if=False,
    **kwargs,
):
    """Helper method to build a resonator with default channels."""
    reson = Resonator(resonator_id, resonator_channel)
    kwargs.pop("fixed_if", None)
    reson.create_pulse_channel(
        ChannelType.measure, *args, fixed_if=measure_fixed_if, **kwargs
    )
    reson.create_pulse_channel(
        ChannelType.acquire, *args, fixed_if=acquire_fixed_if, **kwargs
    )
    return reson


def build_qubit(
    index,
    resonator,
    physical_channel,
    drive_freq,
    second_state_freq=None,
    measure_amp: float = 1.0,
    fixed_drive_if=False,
    qubit_id=None,
):
    """
    Helper method tp build a qubit with assumed default values on the channels. Modelled
    after the live hardware.
    """
    qubit = Qubit(index, resonator, physical_channel, drive_amp=measure_amp, id_=qubit_id)
    qubit.create_pulse_channel(
        ChannelType.drive,
        frequency=drive_freq,
        scale=(1.0e-8 + 0.0j),
        fixed_if=fixed_drive_if,
    )

    qubit.create_pulse_channel(
        ChannelType.second_state, frequency=second_state_freq, scale=(1.0e-8 + 0.0j)
    )

    return qubit


def add_cross_resonance(qubit_one: Qubit, qubit_two: Qubit):
    """Adds cross-resonance couplings and channels between these two qubits."""

    def couple_qubits(qone, qtwo):
        """Builds the channels on qone that will all coupling to qtwo."""
        qone.add_coupled_qubit(qtwo)

        twobit_drive = qtwo.get_drive_channel()
        qone.create_pulse_channel(
            ChannelType.cross_resonance,
            frequency=twobit_drive.frequency,
            scale=(1.0e-7 + 0.0j),
            auxiliary_devices=[qtwo],
        )

        qtwo.create_pulse_channel(
            ChannelType.cross_resonance_cancellation,
            frequency=twobit_drive.frequency,
            scale=(0.0 + 0.0j),
            auxiliary_devices=[qone],
        )

    couple_qubits(qubit_one, qubit_two)
    couple_qubits(qubit_two, qubit_one)


class ChannelType(Enum):
    drive = auto()
    measure = auto()
    second_state = auto()
    cross_resonance = auto()
    cross_resonance_cancellation = auto()
    acquire = auto()
    freq_shift = auto()
    macq = auto()

    def __repr__(self):
        return self.name


class QuantumComponent:
    """
    Base class for any logical object which can act as a target of a quantum action - a
    Qubit or various channels for a simple example.
    """

    def __init__(self, id_, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if id_ is None:
            id_ = ""
        self.id = str(id_)

    def full_id(self):
        return self.id

    def __repr__(self):
        return f"{self.full_id()}"


class Calibratable:
    """
    Allows this object to be loaded as a calibrated object when we get pickled.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_calibrated = False

    def _is_serializable(self, val):
        """
        Checks whether this field value/name should be serialized as a calibration
        value.
        """
        if isinstance(val, (List, tuple)):
            return all([self._is_serializable(val_) for val_ in val])

        if isinstance(val, dict):
            return all([self._is_serializable(val_) for val_ in val.keys()]) and all(
                [self._is_serializable(val_) for val_ in val.values()]
            )

        if isinstance(val, Enum):
            return self._is_serializable(val.value)

        # This checks the field name in JSON is valid, which it'll always be for us.
        return is_picklable("", val)

    def __getstate__(self) -> Dict:
        """
        Returns a dictionary that will be used to pickle the object. By default
        everything that is considered valid by jsonpickle will be added.

        Override this method if you want to explicitly remove things from the
        serialization.
        """
        results = {}
        for key, value in self.__dict__.items():
            if self._is_serializable(value):
                results[key] = value

        return results

    def get_calibration(self):
        return jsonpickle.encode(self, indent=4, context=CyclicRefPickler())

    @staticmethod
    def load_calibration(calibration_string):
        reconstituted = jsonpickle.decode(calibration_string, context=CyclicRefUnpickler())
        if isinstance(reconstituted, str):
            raise ValueError("Loading from calibration string failed. Please regenerate.")

        return reconstituted

    def save_calibration_to_file(self, file_path, use_cwd=False):
        """
        Saves calibration to passed-in file path.
        use_cwd appends the working directory to any non-absolute path, otherwise it'll
        search for common calibration config folders before defaulting to the one in
        source code.
        """
        if use_cwd:
            file_path = os.path.join(os.getcwd(), file_path)

        with open(file_path, "w") as f:
            f.write(self.get_calibration())

    @staticmethod
    def load_calibration_from_file(file_path):
        """
        Looks for this calibration in the passed-in directory or the default calibration
        save location.
        """
        # Check to see if we're just trying to find a file on the working directory.
        cwd_path = os.path.join(os.getcwd(), file_path)
        if os.path.isfile(cwd_path):
            file_path = cwd_path

        with open(file_path, "r") as f:
            return Calibratable.load_calibration(f.read())


class CyclicRefPickler(Pickler):
    """
    Adds reference ID to each object that requires it, for use when re-pickling and
    generating accurate circular references.
    """

    ref_field = "py/obj_ref_id"

    def _flatten_obj_instance(self, obj):
        state = super()._flatten_obj_instance(obj)
        state[self.ref_field] = id(obj)
        return state

    def _log_ref(self, obj):
        is_new = super()._log_ref(obj)
        # We don't want complex' or enums to be referenced. Just create a new one on
        # each instance.
        if isinstance(obj, (complex, Enum, np.number)):
            return True

        if is_new:
            obj_id = id(obj)
            self._objs[obj_id] = obj_id

        return is_new


class FakeList(dict):
    """
    This is a patch-up object because originally the reference-counting mechanism for
    jsonpickle used incrementing ID's and relied upon the graph to be deterministic.

    This isn't the case in some situations, so we want to be able to match up object
    reference by their memory pointer instead. But we don't really want a list a few
    billion elements long for many reasons.

    This simply exposes the methods and functionality jsonpickle relies upon in the
    list, while also allowing us to use very large ints as lookup targets.
    """

    def append(self, val):
        """Appends a value on to the dictionary with key=length."""
        self[len(self)] = val

    def __getitem__(self, item):
        """
        Gets the value, but transforms item-not-found error into an IndexError as
        functionality is triggered upon this exception being thrown (and caught).
        """
        try:
            return super().__getitem__(item)
        except KeyError:
            raise IndexError()


class CyclicRefUnpickler(Unpickler):
    """
    Makes sure cyclic references are picked up correctly. Objects need to be pickled
    with CyclicRefPickler otherwise the tags required for this more precise detection
    won't exist.
    """

    _objs: FakeList

    def reset(self):
        """
        Replace _objs with our own dictionary upon every reset. This is called from
        the constructor.
        """
        super().reset()
        self._objs = FakeList()

    def _restore_object_instance_variables(self, obj, instance):
        # If we see an object with our custom reference pointer replace the incremental
        # number that the dictionary uses to refer to it with the original reference id.
        #
        # The existing code takes care of the reference mapping as long as the py/ref
        # points to the ref_field value. We also strip the custom tag from the
        # dictionary.
        if CyclicRefPickler.ref_field in obj:
            existing_id = obj.pop(CyclicRefPickler.ref_field)
            instance_id = id(instance)
            inc_mapping = self._obj_to_idx[instance_id]
            self._obj_to_idx[instance_id] = existing_id
            del self._objs[inc_mapping]
            self._objs[existing_id] = instance

        return super()._restore_object_instance_variables(obj, instance)


class PhysicalChannel(QuantumComponent, Calibratable):
    def __init__(
        self,
        id_: str,
        sample_time: float,
        baseband: PhysicalBaseband,
        block_size: Optional[int] = None,
        phase_offset: float = 0.0,
        imbalance: float = 1.0,
        acquire_allowed: bool = False,
        pulse_channel_min_frequency: float = 0.0,
        pulse_channel_max_frequency: float = np.inf,
    ):
        super().__init__(id_)
        self.sample_time: float = sample_time
        self.baseband: PhysicalBaseband = baseband
        self.block_size: int = block_size or 1
        self.phase_offset: float = phase_offset
        self.imbalance: float = imbalance

        self.acquire_allowed: bool = acquire_allowed
        self.pulse_channel_min_frequency: float = pulse_channel_min_frequency
        self.pulse_channel_max_frequency: float = pulse_channel_max_frequency

    def create_pulse_channel(
        self,
        id_: str,
        frequency=0.0,
        bias=0.0 + 0.0j,
        scale=1.0 + 0.0j,
        fixed_if: bool = False,
    ):
        pulse_channel = PulseChannel(id_, self, frequency, bias, scale, fixed_if)

        return pulse_channel

    def create_freq_shift_pulse_channel(
        self,
        id_: str,
        frequency=0.0,
        bias=0.0 + 0.0j,
        scale=1.0 + 0.0j,
        amp=0.0,
        active: bool = True,
        fixed_if: bool = False,
    ):
        pulse_channel = FreqShiftPulseChannel(
            id_, self, frequency, bias, scale, amp, active, fixed_if
        )

        return pulse_channel

    @property
    def block_time(self):
        return self.block_size * self.sample_time

    @property
    def baseband_frequency(self):
        return self.baseband.frequency

    @property
    def baseband_if_frequency(self):
        return self.baseband.if_frequency


class PhysicalBaseband(QuantumComponent, Calibratable):
    def __init__(self, id_: str, frequency: float, if_frequency: float = 250e6):
        super().__init__(id_)
        self.frequency: float = frequency
        self.if_frequency: Optional[float] = if_frequency


class PulseChannel(QuantumComponent, Calibratable):
    """Models a pulse channel on a particular device."""

    def __init__(
        self,
        id_: str,
        physical_channel: PhysicalChannel,
        frequency=0.0,
        bias=0.0 + 0.0j,
        scale=1.0 + 0.0j,
        fixed_if: bool = False,
        **kwargs,
    ):
        super().__init__(id_, **kwargs)
        self.physical_channel: PhysicalChannel = physical_channel

        self.frequency: float = frequency
        self.bias: complex = bias
        self.scale: complex = scale

        self.fixed_if: bool = fixed_if

        if frequency < self.min_frequency or frequency > self.max_frequency:
            raise ValueError(
                f"Pulse channel frequency '{frequency}' must be between the bounds "
                f"({self.min_frequency}, {self.max_frequency}) on physical "
                f"channel with id {self.full_id()}."
            )

    @property
    def sample_time(self):
        return self.physical_channel.sample_time

    @property
    def block_size(self):
        return self.physical_channel.block_size

    @property
    def block_time(self):
        return self.physical_channel.block_time

    @property
    def phase_offset(self):
        return self.physical_channel.phase_offset

    @property
    def imbalance(self):
        return self.physical_channel.imbalance

    @property
    def acquire_allowed(self):
        return self.physical_channel.acquire_allowed

    @property
    def baseband_frequency(self):
        return self.physical_channel.baseband_frequency

    @property
    def baseband_if_frequency(self):
        return self.physical_channel.baseband_if_frequency

    @property
    def physical_channel_id(self):
        return self.physical_channel.full_id()

    @property
    def min_frequency(self):
        return self.physical_channel.pulse_channel_min_frequency

    @property
    def max_frequency(self):
        return self.physical_channel.pulse_channel_max_frequency

    def partial_id(self):
        return self.id

    def full_id(self):
        return self.physical_channel_id + "." + self.partial_id()

    def __eq__(self, other):
        if not isinstance(other, PulseChannel):
            return False

        return self.full_id() == other.full_id()

    def __hash__(self):
        return hash(self.full_id())


class FreqShiftPulseChannel(PulseChannel):
    def __init__(
        self,
        id_: str,
        physical_channel: PhysicalChannel,
        frequency=0.0,
        bias=0.0 + 0.0j,
        scale=1.0 + 0.0j,
        amp=0.0,
        active: bool = True,
        fixed_if: bool = False,
        **kwargs,
    ):
        super().__init__(id_, physical_channel, frequency, bias, scale, fixed_if, **kwargs)
        self.amp: float = amp
        self.active: bool = active


class QubitCoupling(Calibratable):
    def __init__(self, direction, quality=1):
        """
        Direction of coupling stated in a tuple: (4,5) means we have a  4 -> 5 coupling.
        Quality is the quality-level of the coupling.
        """
        super().__init__()
        self.direction = tuple(direction)
        self.quality = 1 if quality < 1 else quality

    def __repr__(self):
        return f"{str(self.direction)} @{self.quality}"


class PulseChannelView(PulseChannel):
    """
    Each quantum device will have a unique view of a PulseChannel, which this class
    helps encapsulate. For example, a PulseChannel may be the driving channel for one
    qubit but to another qubit the same PulseChannel might be driving its second state.
    We need to be able to share pulse channel instances with different usages in this
    particular case.

    Functionally this acts as an opaque wrapper to PulseChannel, forwarding all calls to
    the wrapped object.
    """

    # noinspection PyMissingConstructor
    def __init__(
        self,
        pulse_channel: PulseChannel,
        channel_type: ChannelType,
        auxiliary_devices: List[QuantumDevice] = None,
    ):
        """
        pulse_channel: The PulseChannel this device is viewing.
        channel_type: How this devices intends the PulseChannel to be used.
        auxiliary_devices: Any extra devices this PulseChannel could be effecting except
                           the current one.
                           For example in cross resonance pulses.
        """
        # We aren't calling super().__init__ on purpose. Inheriting is purely for show
        # (and typing).
        if auxiliary_devices is None:
            auxiliary_devices = []
        self.pulse_channel: PulseChannel = pulse_channel
        self.channel_type: ChannelType = channel_type
        self.auxiliary_devices: List[QuantumDevice] = auxiliary_devices
        if not any(self._pulse_channel_attributes):
            self._pulse_channel_attributes = {
                val
                for val in dir(pulse_channel)
                if not val.startswith("__") and not val.endswith("__")
            }

            if any(set(vars(self)).intersection(self._pulse_channel_attributes)):
                raise ValueError(
                    "Pulse channel has attributes that will shadow wrapping class."
                )

    _pulse_channel_attributes: Set[str] = set()

    def __getattr__(self, item):
        if item in self._pulse_channel_attributes:
            return getattr(self.pulse_channel, item)
        return super().__getattribute__(item)

    def __setattr__(self, key, value):
        if key in self._pulse_channel_attributes:
            return setattr(self.pulse_channel, key, value)
        return super().__setattr__(key, value)


class QuantumDevice(QuantumComponent, Calibratable):
    """A physical device whose main form of operation involves pulse channels."""

    multi_device_pulse_channel_types = (
        ChannelType.cross_resonance,
        ChannelType.cross_resonance_cancellation,
    )

    def __init__(
        self,
        id_: str,
        physical_channel: PhysicalChannel,
        measure_device: QuantumDevice = None,
    ):
        super().__init__(id_)
        self.measure_device: QuantumDevice = measure_device
        self.pulse_channels: Dict[str, Union[PulseChannel, PulseChannelView]] = {}
        self.default_pulse_channel_type = ChannelType.measure
        self.physical_channel: PhysicalChannel = physical_channel

    def create_pulse_channel(
        self,
        channel_type: ChannelType,
        frequency=0.0,
        bias=0.0 + 0.0j,
        scale=1.0 + 0.0j,
        amp=0.0,
        active: bool = True,
        fixed_if: bool = False,
        auxiliary_devices: List[QuantumDevice] = None,
        id_: str = None,
    ):
        if auxiliary_devices is None:
            auxiliary_devices = []
        if id_ is None:
            id_ = self._create_pulse_channel_id(channel_type, [self] + auxiliary_devices)
        if channel_type == ChannelType.freq_shift:
            pulse_channel = self.physical_channel.create_freq_shift_pulse_channel(
                id_, frequency, bias, scale, amp, active, fixed_if
            )
        else:
            pulse_channel = self.physical_channel.create_pulse_channel(
                id_, frequency, bias, scale, fixed_if
            )
        self.add_pulse_channel(pulse_channel, channel_type, auxiliary_devices)

        return pulse_channel

    def _create_pulse_channel_id(
        self, channel_type: ChannelType, auxiliary_devices: List[QuantumDevice]
    ):
        if (
            channel_type in self.multi_device_pulse_channel_types
            and len(auxiliary_devices) == 0
        ):
            raise ValueError(
                f"Channel type {channel_type.name} requires at least one "
                "auxillary_device"
            )
        return ".".join(
            [str(x.full_id()) for x in (auxiliary_devices)] + [channel_type.name]
        )

    def add_pulse_channel(
        self,
        pulse_channel: PulseChannel,
        channel_type: ChannelType,
        auxiliary_devices: List[QuantumDevice] = None,
    ):
        if auxiliary_devices is None:
            auxiliary_devices = []

        if pulse_channel.physical_channel != self.physical_channel:
            raise ValueError(
                "Pulse channel has physical channel and device physical channel must "
                f"be equal. Device {self} has physical channel {self.physical_channel} "
                "while pulse channel has physical channel "
                f"{pulse_channel.physical_channel}"
            )

        id_ = self._create_pulse_channel_id(channel_type, auxiliary_devices)
        if id_ in self.pulse_channels:
            raise KeyError(f"Pulse channel with id '{id_}' already exists.")

        self.pulse_channels[id_] = PulseChannelView(
            pulse_channel, channel_type, auxiliary_devices
        )

    PTType = TypeVar("PTType", bound=PulseChannel, covariant=True)

    def get_pulse_channel(
        self,
        channel_type: ChannelType = None,
        auxiliary_devices: List[QuantumDevice] = None,
    ) -> PTType:
        if channel_type is None:
            channel_type = self.default_pulse_channel_type
        if auxiliary_devices is None:
            auxiliary_devices = []
        if not isinstance(auxiliary_devices, List):
            auxiliary_devices = [auxiliary_devices]

        id_ = self._create_pulse_channel_id(channel_type, auxiliary_devices)
        if id_ not in self.pulse_channels:
            raise KeyError(
                f"Pulse channel with stored as '{id_}' not found on device '{self.id}'."
            )

        return self.pulse_channels[id_]

    def get_default_pulse_channel(self):
        return self.get_pulse_channel(self.default_pulse_channel_type)

    def get_auxiliary_devices(self, pulse_channel: PulseChannel):
        if isinstance(pulse_channel, PulseChannelView):
            return pulse_channel.auxiliary_devices
        return []


class Resonator(QuantumDevice):
    """Models a resonator on a chip. Can be connected to multiple qubits."""

    def get_measure_channel(self) -> PulseChannel:
        return self.get_pulse_channel(ChannelType.measure)

    def get_acquire_channel(self) -> PulseChannel:
        return self.get_pulse_channel(ChannelType.acquire)


class Qubit(QuantumDevice):
    """
    Class modelling our superconducting qubit and holds all information relating to
    them.
    """

    measure_device: Resonator

    def __init__(
        self,
        index: int,
        resonator: Resonator,
        physical_channel: PhysicalChannel,
        coupled_qubits: List[Qubit] = None,
        drive_amp: float = 1.0,
        id_=None,
    ):
        super().__init__(id_ or f"Q{index}", physical_channel, resonator)
        self.index = index
        self.coupled_qubits: List[Qubit] = []
        self.mean_z_map_args = [1.0, 0.0]
        self.discriminator = [0.0]
        self.pulse_hw_zx_pi_4 = dict()
        self.default_pulse_channel_type = ChannelType.drive

        for qubit in coupled_qubits or []:
            self.add_coupled_qubit(qubit)

        self.pulse_hw_x_pi_2 = {
            "shape": PulseShapeType.GAUSSIAN,
            "width": 100e-9,
            "rise": 1.0 / 3.0,
            "amp": 0.25 / (100e-9 * 1.0 / 3.0 * np.pi**0.5),
            "drag": 0.0,
            "phase": 0.0,
        }

        self.pulse_measure = {
            "shape": PulseShapeType.SQUARE,
            "width": 1.0e-6,
            "amp": drive_amp,
            "amp_setup": 0.0,
            "rise": 0.0,
        }

        self.measure_acquire = {
            "delay": 180e-9,
            "sync": True,
            "width": 1e-6,
            "weights": None,
            "use_weights": False,
        }

    def add_coupled_qubit(self, qubit: Qubit):
        if qubit is None:
            return

        if qubit not in self.coupled_qubits:
            self.coupled_qubits.append(qubit)

        if qubit.full_id() not in self.pulse_hw_zx_pi_4:
            self.pulse_hw_zx_pi_4[qubit.full_id()] = {
                "shape": PulseShapeType.SOFT_SQUARE,
                "width": 125e-9,
                "rise": 10e-9,
                "amp": 1e6,
                "drag": 0.0,
                "phase": 0.0,
            }

    def get_drive_channel(self):
        return self.get_pulse_channel(ChannelType.drive)

    def get_measure_channel(self) -> PulseChannel:
        return self.measure_device.get_measure_channel()

    def get_acquire_channel(self) -> PulseChannel:
        return self.measure_device.get_acquire_channel()

    def get_second_state_channel(self) -> PulseChannel:
        return self.get_pulse_channel(ChannelType.second_state)

    def get_freq_shift_channel(self) -> PulseChannel:
        return self.get_pulse_channel(ChannelType.freq_shift)

    def get_cross_resonance_channel(self, linked_qubits: List[Qubit]) -> PulseChannel:
        return self.get_pulse_channel(ChannelType.cross_resonance, linked_qubits)

    def get_cross_resonance_cancellation_channel(
        self, linked_qubits: List[Qubit]
    ) -> PulseChannel:
        return self.get_pulse_channel(
            ChannelType.cross_resonance_cancellation, linked_qubits
        )

    def get_all_channels(self):
        """
        Returns all channels associated with this qubit, including resonator channel and
        other auxiliary devices that act as if they are on this object.
        """
        return [
            *self.pulse_channels.values(),
            self.get_measure_channel(),
            self.get_acquire_channel(),
        ]


class PulseShapeType(Enum):
    SQUARE = "square"
    GAUSSIAN = "gaussian"
    SOFT_SQUARE = "soft_square"
    BLACKMAN = "blackman"
    SETUP_HOLD = "setup_hold"
    SOFTER_SQUARE = "softer_square"
    EXTRA_SOFT_SQUARE = "extra_soft_square"
    SOFTER_GAUSSIAN = "softer_gaussian"
    ROUNDED_SQUARE = "rounded_square"
    GAUSSIAN_DRAG = "gaussian_drag"
    GAUSSIAN_ZERO_EDGE = "gaussian_zero_edge"
    GAUSSIAN_SQUARE = "gaussian_square"
    SECH = "sech"
    SIN = "sin"
    COS = "cos"

    def __repr__(self):
        return self.value


def _strip_aliases(key: str, build_unaliased=False):
    if "<" in key:
        lhs = key.split("<")
        if len(lhs) > 2:
            raise ValueError(
                f"Alias {key} has more than 1 angle bracket. "
                "Nested or multiple alias' are unsupported."
            )

        rhs = lhs[1].split(">")
        if len(rhs) != 2:
            raise ValueError(
                f"Multiple (or no) closing angle brackets detected in for alias {key}."
            )

        if build_unaliased:
            return lhs[0] + rhs[1]

        return rhs[0]

    return key


MaxPulseLength = 1e-3  # seconds



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
from typing import Dict, Union

from compiler_config.config import CalibrationArguments

from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.runtime import QuantumRuntime, RemoteCalibration, get_runtime

RuntimeOrModel = Union[QuantumHardwareModel, QuantumRuntime]


def find_calibration(args: "CalibrationArguments"):
    calibration = None
    for cali in _standard_calibrations.values():
        if cali.arguments_type() == args.__class__:
            calibration = cali
            break

    if calibration is None:
        raise ValueError(f"Calibration with an argument of {str(args)} couldn't be found.")

    return calibration


def run_calibration(runtime: RuntimeOrModel, args: "CalibrationArguments"):
    calibration = find_calibration(args)

    # If we're just a basic model, use auto-detection.
    if isinstance(runtime, QuantumHardwareModel):
        runtime = get_runtime(runtime)

    calibration.run(runtime.model, runtime, args)


_standard_calibrations: Dict[str, "BuiltinRemoteCalibration"] = dict()


class CustomCalibration(RemoteCalibration):
    """Bespoke calibration built by external users."""

    pass


class BuiltinRemoteCalibration(RemoteCalibration):
    """
    Identifier for any built-in calibrations. Enables automatic running of calibrations
    remotely for any class that inherits this.
    """

    def __init_subclass__(cls, **kwargs):
        _standard_calibrations[cls.__name__] = cls()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
from abc import abstractmethod
from typing import Dict, Union

from qat.purr.compiler.devices import Calibratable, PhysicalBaseband, PhysicalChannel
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class Instrument(Calibratable):
    """
    An interface for any live instrument. It requires a unique address (IP address, USB VISA address, etc.).
    It is derived from Calibratable so the instruments will be saved in the calibration files. To avoid saving
    driver specific data in the calibration files, the actual drivers should be a property of this object, so the
    calibration will skip it.
    """

    def __init__(self, address, id_=None):
        super().__init__()
        self.id = id_ if id_ else address
        self.address = address
        self._driver = None
        self.is_connected = False

    def connect(self):
        if self.is_connected:
            self.disconnect()
        self.is_connected = True
        log.info(f"{type(self).__name__} with ID {self.id} connected")
        return self.is_connected

    def close(self):
        return self.disconnect()

    def disconnect(self):
        if self.driver is not None:
            try:
                self.driver.close()
                self._driver = None
                self.is_connected = False
                log.info(f"{type(self).__name__} with ID {self.id} disconnected")
            except BaseException as e:
                log.warning(
                    f"Failed to close instrument at: {self.address} ID: {self.id}\n{str(e)}"
                )
        return self.is_connected

    @property
    def driver(self):
        return self._driver

    @driver.setter
    def driver(self, obj):
        if self._driver is not None:
            self.disconnect()
        self._driver = obj

    def __getstate__(self) -> Dict:
        results = super(Instrument, self).__getstate__()
        results["_driver"] = None
        return results

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._driver = None
        self.is_connected = False

    def __str__(self):
        return f"{self.__class__.__name__}(id={self.id}, address={self.address})"


class ControlHardwareChannel(PhysicalChannel):
    """
    Wrapper over a PhysicalChannel, that maps to a live instrument channel. This (and
    derived) object should contain hardware specific information.
    """

    def __init__(
        self, id_, hardware_id, dcbiaschannel_pair, switch_ch=None, *args, **kwargs
    ):
        super().__init__(id_, *args, **kwargs)
        self.hardware_id = hardware_id
        self.dcbiaschannel_pair: Dict[str, DCBiasChannel] = dcbiaschannel_pair
        self.switch_ch = switch_ch


class ControlHardware(Instrument):
    """
    The actual control hardware. For now, there is only one per LiveHardwareModel. It
    contains all the physical channels, since in most of the cases, you interact with
    the control unit rather than with the individual channels.
    """

    def __init__(self, id_=None):
        super().__init__(address=None, id_=id_)
        self.channels: Dict[str, ControlHardwareChannel] = {}

    def add_physical_channel(self, *physical_channel: ControlHardwareChannel):
        for physical_channel in physical_channel:
            if physical_channel.id not in self.channels:
                self.channels[physical_channel.id] = physical_channel

        return physical_channel

    def set_data(self, data):
        pass

    def start_playback(self, repetitions, repetition_time):
        pass


class LivePhysicalBaseband(PhysicalBaseband):
    """
    A wrapper over the PhysicalBaseband, that connects to a live instrument.
    """

    def __init__(
        self,
        id_,
        frequency,
        if_frequency,
        instrument: Instrument = None,
        channel_idx=None,
    ):
        self.instrument = instrument
        self.instrument_id = instrument.id if instrument is not None else None
        self.channel_idx = channel_idx
        super().__init__(id_, frequency, if_frequency)

    def connect_to_instrument(self):
        """
        Delayed connection to an instrument, in order to allow instantiating a
        LivePhysicalBaseband without having the live instrument ready.
        """
        if self.instrument is None:
            raise ValueError(f"Cannot connect to instrument '{self.instrument}'!")
        if not self.instrument.is_connected:
            self.instrument.connect()

    @abstractmethod
    def turn_on(self):
        pass

    @abstractmethod
    def turn_off(self):
        pass

    @abstractmethod
    def get_output_state(self):
        pass

    def set_frequency(self, value):
        self.frequency = value


class DCBiasChannel(Calibratable):
    """
    This is generic DC Bias Channel class, It would ONLY accept DC bias card as
    instrument which needs to have get_voltage and set_voltage function.
    """

    def __init__(
        self,
        channel_idx=None,
        bias_value: Union[float, int] = 0,
        instrument: Instrument = None,
    ):
        super().__init__()
        self.channel_idx = channel_idx
        self.instrument = instrument
        self._bias_value = bias_value
        self.max_value = 1
        self.min_value = -1

    def connect_to_instrument(self):
        if self.instrument is None:
            raise ValueError(
                "Cannot connect to instrument 'None', please define/add a DC bias Card!"
            )
        if not self.instrument.is_connected:
            self.instrument.connect()

    @property
    def bias_value(self):
        return self._bias_value

    @bias_value.setter
    def bias_value(self, value):
        self._bias_value = value



from typing import Optional

from pydantic import Field, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class QatConfig(BaseSettings):
    """
    Full settings for a single job. Allows environment variables to be overridden by direct assignment.

    >>> import os
    >>> os.environ["QAT_MAX_REPEATS_LIMIT"] = "654321"
    >>> QatConfig()
    QatConfig(MAX_REPEATS_LIMIT=654321, DISABLE_PULSE_DURATION_LIMITS=False)
    >>> QatConfig(MAX_REPEATS_LIMIT=123)
    QatConfig(MAX_REPEATS_LIMIT=123, DISABLE_PULSE_DURATION_LIMITS=False)

    >>> qatconfig = QatConfig()
    >>> qatconfig.MAX_REPEATS_LIMIT = 16000
    >>> qatconfig
    QatConfig(MAX_REPEATS_LIMIT=16000, DISABLE_PULSE_DURATION_LIMITS=False)

    >>> QatConfig(MAX_REPEATS_LIMIT=100.5) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    pydantic_core._pydantic_core.ValidationError
    ...
    Input should be a valid integer, got a number with a fractional part
    """

    model_config = SettingsConfigDict(env_prefix="QAT_", validate_assignment=True)
    MAX_REPEATS_LIMIT: Optional[int] = Field(gt=0, default=100_000)
    """Max number of repeats / shots to be performed in a single job."""
    DISABLE_PULSE_DURATION_LIMITS: bool = False
    """Flag to disable the lower and upper pulse duration limits. 
    Only needs to be set to True for calibration purposes."""

    @field_validator("DISABLE_PULSE_DURATION_LIMITS")
    def check_disable_pulse_duration_limits(cls, DISABLE_PULSE_DURATION_LIMITS):
        if DISABLE_PULSE_DURATION_LIMITS:
            log.warning(
                "Disabled check for pulse duration limits, which should ideally only be used for calibration purposes."
            )
        return DISABLE_PULSE_DURATION_LIMITS


qatconfig = QatConfig()

if __name__ == "__main__":
    import doctest

    doctest.testmod()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
from typing import List, Union

from compiler_config.config import InlineResultsProcessing

from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.execution import InstructionExecutionEngine
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import Variable
from qat.purr.compiler.runtime import get_builder
from qat.purr.utils.logging_utils import log_duration

qir_available = True
try:
    from pyqir import (
        Call,
        Constant,
        Context,
        FloatConstant,
        IntConstant,
        Module,
        extract_byte_string,
        is_entry_point,
        qubit_id,
        result_id,
    )
except ImportError:
    qir_available = False


class QIRParser:
    def __init__(
        self,
        hardware: Union[QuantumHardwareModel, InstructionExecutionEngine],
        builder=None,
    ):
        if isinstance(hardware, InstructionExecutionEngine):
            hardware = hardware.model

        self.hardware: QuantumHardwareModel = hardware
        self.builder: InstructionBuilder = builder or get_builder(hardware)
        self.results_format = InlineResultsProcessing.Program
        self.result_variables = []

    def _get_qubit(self, id_: int):
        return self.hardware.qubits[id_]

    def ccx(self, control1, control2, target):
        self.builder.ccnot(
            [self._get_qubit(control1), self._get_qubit(control2)],
            target,
        )

    def cx(self, control: int, target: int):
        self.builder.cnot(self._get_qubit(control), self._get_qubit(target))

    def cz(self, control: int, target: int):
        self.builder.cZ(self._get_qubit(control), self._get_qubit(target))

    def h(self, target: int):
        self.builder.had(self._get_qubit(target))

    def mz(self, qubit: int, target: int):
        result_var = str(target)
        self.builder.measure_single_shot_z(
            self._get_qubit(qubit), output_variable=result_var
        )
        self.builder.results_processing(result_var, self.results_format)

    def reset(self, target: int):
        self.builder.reset(self._get_qubit(target))

    def rx(self, theta: float, qubit: int):
        self.builder.X(self._get_qubit(qubit), theta)

    def ry(self, theta: float, qubit: int):
        self.builder.Y(self._get_qubit(qubit), theta)

    def rz(self, theta: float, qubit: int):
        self.builder.Z(self._get_qubit(qubit), theta)

    def s(self, qubit: int):
        self.builder.S(self._get_qubit(qubit))

    def s_adj(self, qubit: int):
        self.builder.Sdg(self._get_qubit(qubit))

    def t(self, qubit: int):
        self.builder.T(self._get_qubit(qubit))

    def t_adj(self, qubit: int):
        self.builder.Tdg(self._get_qubit(qubit))

    def x(self, qubit: int):
        self.builder.X(self._get_qubit(qubit))

    def y(self, qubit: int):
        self.builder.Y(self._get_qubit(qubit))

    def z(self, qubit: int):
        self.builder.Z(self._get_qubit(qubit))

    def process_instructions(self, instructions):
        for inst in instructions:
            if isinstance(inst, Call):

                def throw_on_invalid_args(actual_args, expected_args):
                    if actual_args != expected_args:
                        raise ValueError(
                            f"{intrinsic_name} has {actual_args} arguments, "
                            f"expected {expected_args}."
                        )

                args: List[Constant] = inst.args
                intrinsic_name = inst.callee.name
                if intrinsic_name in (
                    "__quantum__qis__ccx__body",
                    "__quantum__qis__ccnot__body",
                ):
                    throw_on_invalid_args(len(args), 3)
                    self.ccx(qubit_id(args[0]), qubit_id(args[1]), qubit_id(args[2]))
                elif intrinsic_name in (
                    "__quantum__qis__cnot__body",
                    "__quantum__qis__cx__body",
                ):
                    throw_on_invalid_args(len(args), 2)
                    self.cx(qubit_id(args[0]), qubit_id(args[1]))
                elif intrinsic_name == "__quantum__qis__cz__body":
                    throw_on_invalid_args(len(args), 2)
                    self.cz(qubit_id(args[0]), qubit_id(args[1]))
                elif intrinsic_name == "__quantum__qis__h__body":
                    throw_on_invalid_args(len(args), 1)
                    self.h(qubit_id(args[0]))
                elif intrinsic_name == "__quantum__qis__mz__body":
                    throw_on_invalid_args(len(args), 2)
                    self.mz(qubit_id(args[0]), result_id(args[1]))
                elif intrinsic_name == "__quantum__qis__reset__body":
                    throw_on_invalid_args(len(args), 1)
                    self.reset(qubit_id(args[0]))
                elif intrinsic_name == "__quantum__qis__rx__body":
                    throw_on_invalid_args(len(args), 2)
                    radii: Union[IntConstant, FloatConstant] = args[0]
                    self.rx(radii.value, qubit_id(args[1]))
                elif intrinsic_name == "__quantum__qis__ry__body":
                    throw_on_invalid_args(len(args), 2)
                    radii: Union[IntConstant, FloatConstant] = args[0]
                    self.ry(radii.value, qubit_id(args[1]))
                elif intrinsic_name == "__quantum__qis__rz__body":
                    throw_on_invalid_args(len(args), 2)
                    radii: Union[IntConstant, FloatConstant] = args[0]
                    self.rz(radii.value, qubit_id(args[1]))
                elif intrinsic_name == "__quantum__qis__s__body":
                    throw_on_invalid_args(len(args), 1)
                    self.s(qubit_id(args[0]))
                elif intrinsic_name == "__quantum__qis__s_adj":
                    throw_on_invalid_args(len(args), 1)
                    self.s_adj(qubit_id(args[0]))
                elif intrinsic_name == "__quantum__qis__t__body":
                    throw_on_invalid_args(len(args), 1)
                    self.t(qubit_id(args[0]))
                elif intrinsic_name == "__quantum__qis__t__adj":
                    throw_on_invalid_args(len(args), 1)
                    self.t_adj(qubit_id(args[0]))
                elif intrinsic_name == "__quantum__qis__x__body":
                    throw_on_invalid_args(len(args), 1)
                    self.x(qubit_id(args[0]))
                elif intrinsic_name == "__quantum__qis__y__body":
                    throw_on_invalid_args(len(args), 1)
                    self.y(qubit_id(args[0]))
                elif intrinsic_name == "__quantum__qis__z__body":
                    throw_on_invalid_args(len(args), 1)
                    self.z(qubit_id(args[0]))
                elif intrinsic_name == "__quantum__rt__initialize":
                    pass
                elif intrinsic_name == "__quantum__rt__tuple_record_output":
                    pass
                elif intrinsic_name == "__quantum__rt__array_record_output":
                    pass
                elif intrinsic_name == "__quantum__rt__result_record_output":
                    throw_on_invalid_args(len(args), 2)
                    res = result_id(args[0])
                    label_ptr = args[1]
                    label = ""
                    if (not label_ptr.is_null) and (
                        byte_string := extract_byte_string(label_ptr)
                    ):
                        label = byte_string.decode("utf-8").rstrip("\x00")
                    self.result_variables.append((Variable(str(res)), label))

    def parse(self, qir_file: str):
        if not qir_available:
            raise RuntimeError("QIR parser unavailable.")

        with log_duration("QIR parsing completed, took {} seconds."):
            if qir_file.endswith(".bc"):
                with open(qir_file, "rb") as f:
                    mod = Module.from_bitcode(Context(), f.read())
            elif qir_file.endswith(".ll"):
                with open(qir_file) as f:
                    mod = Module.from_ir(Context(), f.read())

            entry_point = next(
                iter(filter(lambda fnc: is_entry_point(fnc), mod.functions)), None
            )
            if entry_point is None:
                raise ValueError("Entry point unable to be found in QIR file.")

            self.process_instructions(
                [inst for bb in entry_point.basic_blocks for inst in bb.instructions]
            )

            if any(self.result_variables):
                potential_names = [
                    val[1] for val in self.result_variables if len(val[1] or "") != 0
                ]
                if not any(potential_names):
                    result_name = Variable.generate_name()
                else:
                    result_name = "_".join(potential_names)

                self.builder.assign(result_name, [val[0] for val in self.result_variables])
                self.builder.returns(result_name)
            else:
                self.builder.returns()

            complete_builder = self.builder
            self.builder = get_builder(self.hardware)
            return complete_builder



import json
import os
from dataclasses import asdict
from datetime import datetime
from functools import reduce
from typing import Dict, List

import numpy as np
import regex
from qblox_instruments import Cluster, DummyBinnedAcquisitionData, DummyScopeAcquisitionData
from qblox_instruments.qcodes_drivers.module import Module
from qblox_instruments.qcodes_drivers.sequencer import Sequencer

from qat.purr.backends.live_devices import ControlHardware, Instrument, LivePhysicalBaseband
from qat.purr.backends.qblox.codegen import QbloxPackage
from qat.purr.backends.qblox.config import (
    QbloxConfig,
    QcmConfigHelper,
    QcmRfConfigHelper,
    QrmConfigHelper,
    QrmRfConfigHelper,
)
from qat.purr.backends.qblox.constants import Constants
from qat.purr.backends.qblox.ir import Sequence
from qat.purr.backends.qblox.visualisation import plot_acquisitions, plot_packages
from qat.purr.compiler.devices import (
    ChannelType,
    PhysicalChannel,
    PulseChannel,
    Qubit,
    Resonator,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class QbloxResonator(Resonator):
    """
    A hack around QBlox acquisition to use the same pulse channel
    for MeasurePulse and Acquire instructions.
    """

    def get_measure_channel(self) -> PulseChannel:
        return self.get_pulse_channel(ChannelType.macq)

    def get_acquire_channel(self) -> PulseChannel:
        return self.get_pulse_channel(ChannelType.macq)


class QbloxPhysicalBaseband(LivePhysicalBaseband):
    def __init__(
        self,
        id_,
        frequency,
        if_frequency,
        instrument: Instrument,
        slot_idx: int,
        config: QbloxConfig,
    ):
        super().__init__(id_, frequency, if_frequency, instrument)
        self.slot_idx = slot_idx
        self.config = config  # TBD, is it worth flattening ?


class QbloxPhysicalChannel(PhysicalChannel):
    baseband: QbloxPhysicalBaseband

    @property
    def slot_idx(self):
        return self.baseband.slot_idx

    @property
    def config(self):
        return self.baseband.config

    def build_resonator(self, resonator_id, *args, **kwargs):
        """Helper method to build a resonator with default channels."""
        kwargs.pop("fixed_if", None)
        reson = QbloxResonator(resonator_id, self)
        reson.create_pulse_channel(ChannelType.macq, *args, fixed_if=False, **kwargs)
        reson.default_pulse_channel_type = ChannelType.macq
        return reson

    def build_qubit(
        self,
        index,
        resonator,
        drive_freq,
        second_state_freq=None,
        channel_scale=(1.0e-8 + 0.0j),
        measure_amp: float = 1.0,
        fixed_drive_if=False,
        qubit_id=None,
    ):
        """
        Helper method tp build a qubit with assumed default values on the channels. Modelled after the live hardware.
        """
        qubit = Qubit(index, resonator, self, drive_amp=measure_amp, id_=qubit_id)
        qubit.create_pulse_channel(
            ChannelType.drive,
            frequency=drive_freq,
            scale=channel_scale,
            fixed_if=False,
        )

        qubit.create_pulse_channel(
            ChannelType.second_state, frequency=second_state_freq, scale=channel_scale
        )

        qubit.measure_acquire["delay"] = 151e-9  # TOF

        return qubit


class QbloxControlHardware(ControlHardware):
    _driver: Cluster

    def __init__(
        self,
        dev_id: str = None,
        name: str = None,
        address: str = None,
        dummy_cfg: Dict = None,
    ):
        super().__init__(id_=dev_id or os.environ.get("QBLOX_DEV_ID"))
        self.name = name or os.environ.get("QBLOX_DEV_NAME")
        self.address = address or os.environ.get("QBLOX_DEV_IP")
        self.dummy_cfg = dummy_cfg
        self.dump_sequence = False
        self.plot_packages = False
        self.plot_acquisitions = False
        self._resources: Dict[Module, Dict[PulseChannel, Sequencer]] = {}

    def allocate_resources(self, packages: List[QbloxPackage]):
        for pkg in packages:
            target = pkg.target
            module_id = target.physical_channel.slot_idx - 1  # slot_idx is in range [1..20]
            module: Module = self._driver.modules[module_id]
            allocations = self._resources.setdefault(module, {})
            if target not in allocations:
                total = set(target.physical_channel.config.sequencers.keys())
                allocated = set([seq.seq_idx for seq in allocations.values()])

                available = total - allocated
                if not available:
                    raise ValueError(f"No more available sequencers on module {module}")
                sequencer: Sequencer = module.sequencers[next(iter(available))]
                allocations[target] = sequencer

        return self._resources

    def _delete_acquisitions(self, sequencer):
        sequencer.delete_acquisition_data(all=True)

    def _prepare_config(self, package: QbloxPackage, sequencer: Sequencer):
        if package.target.fixed_if:  # NCO freq constant
            nco_freq = package.target.baseband_if_frequency
            lo_freq = package.target.frequency - nco_freq
        else:  # LO freq constant
            lo_freq = package.target.baseband_frequency
            nco_freq = package.target.frequency - lo_freq

        qblox_config = package.target.physical_channel.config

        module_config = qblox_config.module
        if module_config.lo.out0_en:
            module_config.lo.out0_freq = lo_freq
        if module_config.lo.out1_en:
            module_config.lo.out1_freq = lo_freq
        if module_config.lo.out0_in0_en:
            module_config.lo.out0_in0_freq = lo_freq

        # Sequencer config from the HwM
        hwm_seq_config = qblox_config.sequencers[sequencer.seq_idx]
        # Sequencer config from QAT IR
        pkg_seq_config = package.sequencer_config
        hwm_seq_config.nco.freq = nco_freq
        hwm_seq_config.square_weight_acq.integration_length = (
            pkg_seq_config.square_weight_acq.integration_length
        )

        return qblox_config

    def _reset_io(self, modules=None):
        # TODO - Qblox bug: Hard reset clutters sequencer connections with conflicting defaults
        # TODO - This is a temporary workaround until Qblox fixes the issue

        modules = modules or self._driver.get_connected_modules().values()

        for m in modules:
            log.info(f"Resetting sequencer connections for module {m.slot_idx}")
            m.disconnect_outputs()
            if m.is_qrm_type:
                m.disconnect_inputs()

    def connect(self):
        if self._driver is None or not Cluster.is_valid(self._driver):
            self._driver: Cluster = Cluster(
                name=self.name,
                identifier=self.address,
                dummy_cfg=self.dummy_cfg if self.address is None else None,
            )
            self._driver.reset()
            self._reset_io()

        log.info(self._driver.get_system_status())
        self.is_connected = True

    def disconnect(self):
        if self._driver is not None:
            try:
                self._driver.close()
                self.is_connected = False
            except BaseException as e:
                log.warning(
                    f"Failed to close instrument ID: {self.id} at: {self.address}\n{str(e)}"
                )

    def install(self, package: QbloxPackage, module: Module, sequencer: Sequencer):
        """
        Installs and configures the package on the given (module, sequencer) resource.
        Nominally follows resource allocation, but also serves for adhoc package installations
        that aren't necessarily tied to any automatic resource allocation.
        """

        log.debug(f"Configuring module {module}, sequencer {sequencer}")
        config = self._prepare_config(package, sequencer)
        if module.is_qcm_type:
            if module.is_rf_type:
                QcmRfConfigHelper(config).configure(module, sequencer)
            else:
                QcmConfigHelper(config).configure(module, sequencer)
        elif module.is_qrm_type:
            if module.is_rf_type:
                QrmRfConfigHelper(config).configure(module, sequencer)
            else:
                QrmConfigHelper(config).configure(module, sequencer)

        sequence = asdict(package.sequence)
        log.debug(f"Uploading sequence to {module}, sequencer {sequencer}")
        sequencer.sequence(sequence)

        return module, sequencer

    def set_data(self, qblox_packages: List[QbloxPackage]):
        self._resources.clear()
        self.allocate_resources(qblox_packages)

        if self.plot_packages:
            plot_packages(qblox_packages)

        try:
            for package in qblox_packages:
                module, sequencer = next(
                    (
                        (m, t2s[package.target])
                        for m, t2s in self._resources.items()
                        if m.slot_idx == package.target.physical_channel.slot_idx
                    )
                )

                if self.dump_sequence:
                    filename = f"schedules/target_{package.target.id}_module_{module.slot_idx}_sequencer_{sequencer.seq_idx}_@_{datetime.utcnow().strftime('%m-%d-%Y_%H%M%S')}.json"
                    os.makedirs(os.path.dirname(filename), exist_ok=True)
                    with open(filename, "w") as f:
                        f.write(json.dumps(asdict(package.sequence)))

                self.install(package, module, sequencer)
        except BaseException as e:
            self._reset_io(self._resources.keys())
            raise e

    def start_playback(self, repetitions: int, repetition_time: float):
        if not any(self._resources):
            raise ValueError("No resources allocated. Install packages first")

        results = {}
        try:
            for module, allocations in self._resources.items():
                for target, sequencer in allocations.items():
                    sequencer.sync_en(True)

            for module, allocations in self._resources.items():
                for target, sequencer in allocations.items():
                    sequencer.arm_sequencer()

            for module, allocations in self._resources.items():
                for target, sequencer in allocations.items():
                    sequencer.start_sequencer()

            for module, allocations in self._resources.items():
                if module.is_qrm_type:
                    for target, sequencer in allocations.items():
                        sequencer.get_acquisition_status(timeout=1)
                        acquisitions = sequencer.get_acquisitions()

                        for acq_name, acq in acquisitions.items():
                            sequencer.store_scope_acquisition(acq_name)

                            i = np.array(acq["acquisition"]["bins"]["integration"]["path0"])
                            q = np.array(acq["acquisition"]["bins"]["integration"]["path1"])
                            results[acq_name] = (
                                i + 1j * q
                            ) / sequencer.integration_length_acq()

                        if self.plot_acquisitions:
                            plot_acquisitions(
                                sequencer.get_acquisitions(),  # (re)fetch after store
                                integration_length=sequencer.integration_length_acq(),
                            )

                        self._delete_acquisitions(sequencer)
        finally:
            for module, allocations in self._resources.items():
                for target, sequencer in allocations.items():
                    sequencer.sync_en(False)

                module.disconnect_outputs()
                if module.is_qrm_type:
                    module.disconnect_inputs()

        return results

    def __getstate__(self) -> Dict:
        results = super(QbloxControlHardware, self).__getstate__()
        results["_driver"] = None
        results["_resources"] = {}
        return results

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._driver = None
        self._resources = {}
        self.is_connected = False


class DummyQbloxControlHardware(QbloxControlHardware):
    def _setup_dummy_scope_acq_data(self, module, sequencer: Sequencer, sequence: Sequence):
        if sequence.waveforms:
            dummy_data = reduce(
                lambda a, b: a + b, [a["data"] for a in sequence.waveforms.values()]
            )
            dummy_data = [(z, z) for z in dummy_data / np.linalg.norm(dummy_data)]
        else:
            playback_pattern = regex.compile(
                "set_awg_offs( +)([0-9]+),([0-9]+)\nupd_param( +)([0-9]+)"
            )
            match = next(playback_pattern.finditer(sequence.program))
            i_steps, q_steps = int(match.group(2)), int(match.group(3))
            num_samples = int(match.group(5))
            dummy_data = [
                (
                    i_steps / (Constants.MAX_OFFSET_SIZE // 2),
                    q_steps / (Constants.MAX_OFFSET_SIZE // 2),
                )
            ] * num_samples

        dummy_scope_acquisition_data = DummyScopeAcquisitionData(
            data=dummy_data, out_of_range=(False, False), avg_cnt=(1, 1)
        )
        module.set_dummy_scope_acquisition_data(
            sequencer=sequencer.seq_idx, data=dummy_scope_acquisition_data
        )

    def _setup_dummy_binned_acq_data(
        self, module, sequencer: Sequencer, sequence: Sequence
    ):
        for name, acquisition in sequence.acquisitions.items():
            dummy_data = (np.random.random(), np.random.random())
            dummy_binned_acquisition_data = [
                DummyBinnedAcquisitionData(data=dummy_data, thres=1, avg_cnt=1)
            ] * acquisition["num_bins"]
            module.set_dummy_binned_acquisition_data(
                sequencer=sequencer.seq_idx,
                acq_index_name=name,
                data=dummy_binned_acquisition_data,
            )

    def _delete_acquisitions(self, sequencer):
        sequencer.delete_dummy_scope_acquisition_data()
        sequencer.delete_dummy_binned_acquisition_data()

    def set_data(self, qblox_packages: List[QbloxPackage]):
        super().set_data(qblox_packages)

        # Stage Scope and Acquisition data
        for module, allocations in self._resources.items():
            if module.is_qrm_type:
                for target, sequencer in allocations.items():
                    package = next((pkg for pkg in qblox_packages if pkg.target == target))
                    self._setup_dummy_scope_acq_data(module, sequencer, package.sequence)
                    self._setup_dummy_binned_acq_data(module, sequencer, package.sequence)



from compiler_config.config import CompilerConfig

import qat.purr.compiler.frontends as core_frontends
from qat.purr.backends.calibrations.remote import find_calibration
from qat.purr.backends.qiskit_simulator import QiskitEngine, QiskitHardwareModel
from qat.purr.compiler.interrupt import Interrupt, NullInterrupt
from qat.purr.compiler.runtime import (
    _execute_instructions_with_interrupt,
    execute_instructions,
)


class InterruptableExecutingMixin:
    """
    Support experimental wait interrupt capabilities
    """

    def _execute(
        self,
        hardware,
        compiler_config: CompilerConfig,
        instructions,
        *args,
        interrupt: Interrupt = NullInterrupt(),
        **kwargs,
    ):

        calibrations = [
            find_calibration(arg) for arg in compiler_config.active_calibrations
        ]

        exe_method = _execute_instructions_with_interrupt
        if isinstance(hardware, (QiskitHardwareModel, QiskitEngine)):
            exe_method = execute_instructions

        return exe_method(hardware, instructions, compiler_config, calibrations, interrupt)


class QIRFrontend(InterruptableExecutingMixin, core_frontends.QIRFrontend):
    """
    Static decorated QIR frontend for interruptibility
    """

    pass


class QASMFrontend(InterruptableExecutingMixin, core_frontends.QASMFrontend):
    """
    Static decorated QASM frontend for interruptibility
    """

    pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd

from abc import ABC, abstractmethod
from typing import List, Optional

from compiler_config.config import CompilerConfig

from qat.purr.backends.live import LiveHardwareModel, build_lucy_hardware
from qat.purr.backends.live_devices import ControlHardware
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.emitter import QatFile
from qat.purr.compiler.execution import QuantumExecutionEngine
from qat.purr.compiler.instructions import Instruction, QuantumInstruction
from qat.purr.compiler.interrupt import Interrupt, NullInterrupt
from qat.purr.utils.logger import get_default_logger
from qat.qat import execute

log = get_default_logger()


class QPUVersion:
    def __init__(self, make: str, version: str = None):
        self.make = make
        self.version = version

    def __repr__(self):
        return f"{self.make}-{self.version if self.version is not None else 'latest'}"

    @staticmethod
    def with_version(version: str = None):
        """Creates a QPU version with an empty QPU make. Only used in very special circumstances."""
        return QPUVersion("", version)


def inject_name(cls):
    """
    Decorator to inject class name into the QPU make/version fields
    in our static make/model naming objects.
    """
    for value in vars(cls).values():
        if isinstance(value, QPUVersion):
            if value.make is None or value.make == "":
                value.make = cls.__name__

    return cls


@inject_name
class Lucy:
    Latest = QPUVersion.with_version()


class VerificationModel(LiveHardwareModel):
    def __init__(
        self,
        qpu_version,
        verification_engine: "VerificationEngine",
        control_hardware: ControlHardware = ControlHardware(),
    ):
        super().__init__(control_hardware)
        self.version = qpu_version
        self.verification_engine = verification_engine

    def create_engine(self) -> "VerificationEngine":
        return self.verification_engine(self)


def verify_instructions(builder: InstructionBuilder, qpu_type: QPUVersion):
    """
    Runs instruction verification for the instructions in this builder.

    Only run this on instructions that will go through no more transformations
    before being sent to the driver, otherwise you cannot rely upon its results as
    being accurate. In most situations this will only be fully truthful on instructions that
    have already gone through the entire pipeline.
    """
    model = get_verification_model(qpu_type)
    if model is None:
        raise ValueError(
            f"Cannot verify instructions, {qpu_type} isn't a valid QPU version."
        )

    engine: VerificationEngine = model.get_engine()
    engine.verify_instructions(builder.instructions)


def get_verification_model(qpu_type: QPUVersion) -> Optional[VerificationModel]:
    """
    Get verification model for a particular QPU make and model. Each make has its own class,
    which has a field that is each individual version available for verification.

    For example, if you wanted to verify our Lucy machine, that'd be done with:
    ``
    get_verification_model(Lucy.Latest)
    ``

    Or with a specific version:
    ``
    get_verification_model(Lucy.XY)
    ``
    """
    if not isinstance(qpu_type, QPUVersion):
        raise ValueError(
            f"{qpu_type} is not a QPU version, can't find verification engine."
        )

    if qpu_type.make == Lucy.__name__:
        return build_lucy_hardware(VerificationModel(qpu_type, LucyVerificationEngine))

    return None


class VerificationEngine(QuantumExecutionEngine, ABC):
    def _execute_on_hardware(
        self, sweep_iterator, package: QatFile, interrupt: Interrupt = NullInterrupt()
    ):
        while not sweep_iterator.is_finished():
            sweep_iterator.do_sweep(package.instructions)

            metadata = {"sweep_iteration": sweep_iterator.get_current_sweep_iteration()}
            interrupt.if_triggered(metadata, throw=True)
            self.verify_instructions(package.instructions, package.meta_instructions)

        return dict()

    def _process_results(self, results, qfile: "QatFile"):
        return results

    def _process_assigns(self, results, qfile: "QatFile"):
        return results

    @abstractmethod
    def verify_instructions(self, instructions: List[Instruction], metadata): ...


class LucyVerificationEngine(VerificationEngine):

    max_circuit_duration = 90000e-9

    def verify_instructions(self, instructions: List[QuantumInstruction], metadata):
        timeline = self.create_duration_timeline(instructions)

        pc2samples = {pc: positions[-1].end for pc, positions in timeline.items()}
        durations = {pc: samples * pc.sample_time for pc, samples in pc2samples.items()}

        circuit_duration = max([duration for duration in durations.values()])

        log.debug(f"The circuit duration is {circuit_duration/1e-6} microseconds.")

        if circuit_duration > self.max_circuit_duration:
            raise VerificationError(
                f"Circuit duration exceeds maximum allowed. Duration {circuit_duration}, max: {self.max_circuit_duration}."
            )


def verify_program(program: str, compiler_config: CompilerConfig, qpu_version: QPUVersion):
    model = get_verification_model(qpu_version)
    if model is None:
        raise ValueError("Unable to find model to verify against.")

    return execute(program, model, compiler_config)


class VerificationError(ValueError): ...



from numbers import Number
from typing import Dict, List

from qat.ir.pass_base import TransformPass
from qat.ir.result_base import ResultManager
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import PulseChannel
from qat.purr.compiler.instructions import (
    AcquireMode,
    CustomPulse,
    Instruction,
    PhaseReset,
    PhaseShift,
    PostProcessing,
    PostProcessType,
    ProcessAxis,
    Pulse,
)


class PhaseOptimisation(TransformPass):
    """
    Extracted from QuantumExecutionEngine.optimize()
    """

    def run(self, builder: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        accum_phaseshifts: Dict[PulseChannel, PhaseShift] = {}
        optimized_instructions: List[Instruction] = []
        for instruction in builder.instructions:
            if isinstance(instruction, PhaseShift) and isinstance(
                instruction.phase, Number
            ):
                if accum_phaseshift := accum_phaseshifts.get(instruction.channel, None):
                    accum_phaseshift.phase += instruction.phase
                else:
                    accum_phaseshifts[instruction.channel] = PhaseShift(
                        instruction.channel, instruction.phase
                    )
            elif isinstance(instruction, (Pulse, CustomPulse)):
                quantum_targets = getattr(instruction, "quantum_targets", [])
                if not isinstance(quantum_targets, List):
                    quantum_targets = [quantum_targets]
                for quantum_target in quantum_targets:
                    if quantum_target in accum_phaseshifts:
                        optimized_instructions.append(accum_phaseshifts.pop(quantum_target))
                optimized_instructions.append(instruction)
            elif isinstance(instruction, PhaseReset):
                for channel in instruction.quantum_targets:
                    accum_phaseshifts.pop(channel, None)
                optimized_instructions.append(instruction)
            else:
                optimized_instructions.append(instruction)

        builder.instructions = optimized_instructions


class PostProcessingOptimisation(TransformPass):
    """
    Extracted from LiveDeviceEngine.optimize()
    Better pass name/id ?
    """

    def run(self, builder: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        pp_insts = [val for val in builder.instructions if isinstance(val, PostProcessing)]
        discarded = []
        for pp in pp_insts:
            if pp.acquire.mode == AcquireMode.SCOPE:
                if (
                    pp.process == PostProcessType.MEAN
                    and ProcessAxis.SEQUENCE in pp.axes
                    and len(pp.axes) <= 1
                ):
                    discarded.append(pp)

            elif pp.acquire.mode == AcquireMode.INTEGRATOR:
                if (
                    pp.process == PostProcessType.DOWN_CONVERT
                    and ProcessAxis.TIME in pp.axes
                    and len(pp.axes) <= 1
                ):
                    discarded.append(pp)
                if (
                    pp.process == PostProcessType.MEAN
                    and ProcessAxis.TIME in pp.axes
                    and len(pp.axes) <= 1
                ):
                    discarded.append(pp)
        builder.instructions = [val for val in builder.instructions if val not in discarded]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
from dataclasses import dataclass, field
from enum import Enum, auto
from random import random
from typing import Dict, Iterable, List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import CubicSpline

from qat.purr.compiler.devices import (
    Calibratable,
    ChannelType,
    PhysicalBaseband,
    PhysicalChannel,
    QuantumComponent,
    Qubit,
    Resonator,
)
from qat.purr.compiler.emitter import QatFile
from qat.purr.compiler.instructions import (
    AcquireMode,
    Instruction,
    PostProcessing,
    Repeat,
    Reset,
)
from qat.purr.compiler.interrupt import Interrupt, NullInterrupt
from qat.purr.utils.logger import get_default_logger

from ..compiler.execution import QuantumExecutionEngine, SweepIterator
from ..compiler.hardware_models import QuantumHardwareModel
from ..utils.logging_utils import log_duration
from .utilities import UPCONVERT_SIGN, PositionData, get_axis_map

logger = get_default_logger()

try:
    from qutip import Qobj, basis, create, destroy, mesolve, qeye, tensor

    qutip_available = True
except ModuleNotFoundError:
    qutip_available = False


class ControlType(Enum):
    MEASURE = "measure"
    RESET = "reset"

    def __repr__(self):
        return self.name


@dataclass
class OperatorInfo:
    operator: "Qobj"
    name: str
    qubit_id: int = None

    def __repr__(self):
        return f"Operator({self.name}, {self.qubit_id})"


@dataclass
class Section:
    indices: List[int]
    qubits: List[str]
    control_type: ControlType
    simulation_time: List[float] = None
    drive_hamiltonian: List["Qobj"] = field(default_factory=list)

    def __repr__(self):
        return f"Section({self.indices}, {self.qubits}, {str(self.control_type)})"


class MeasurementStatistics:
    def __init__(self, state, level=0, probability=1, qubits_measured=None, result=None):
        self.state: "Qobj" = state
        self.level: int = level
        self.probability: float = probability
        self.result: str = result
        self.qubits_measured: List[int] = qubits_measured
        self.measurement_statistics: List[MeasurementStatistics] = []
        self.branch_number: Union[int, None] = None
        self.dynamics_results = None
        self.hamiltonian: Union[List["Qobj", float], None] = None
        self.c_ops: Union[List["Qobj"], None] = None
        self.sim_t: Union[List[float], None] = None

    def build_tree(
        self,
        H0,
        c_ops,
        simulation_sections,
        qubit_states,
        reset_operators,
        options=None,
        branch_number=-1,
    ):
        """
        Simulate the dynamics of initial condition using the section of the Hamiltonian
        specified by the level. Perform a measurement on the specified qubits in the
        Hamiltonian section and for each result calculate the probability of it occuring
        and the resultant state. For each result perform a new simulation with the next
        Hamiltonian section to create a branching structure of all possible outcomes for
        a given initial condition and Hamiltonian sections dictionary. Continue until
        all Hamiltonian sections have been simulated.
        """
        if self.level == len(simulation_sections) or self.probability.real <= 0:
            branch_number += 1
            self.branch_number = branch_number
            return branch_number

        section = simulation_sections[self.level]
        control_type = section.control_type
        qubit_indices = section.qubits
        drive_hamiltonian = section.drive_hamiltonian
        sim_t = section.simulation_time

        H = [H0]
        H.extend(drive_hamiltonian)

        if len(drive_hamiltonian) == 0:
            self.dynamics_results = [self.state]
            final_state = self.state
        else:
            result = mesolve(H, self.state, sim_t, c_ops=c_ops, options=options)
            self.dynamics_results = result.states
            final_state = result.states[-1]
        self.hamiltonian = H
        self.c_ops = c_ops
        self.sim_t = sim_t

        if control_type == ControlType.MEASURE:
            empty_state = H0 - H0
            qubit_indices.sort()
            sub_space_size = len(qubit_indices)
            # Create list of sub states from qubits to be measured
            measurement_operators = {
                format(out_state, f"0{sub_space_size}b"): empty_state
                for out_state in range(2**sub_space_size)
            }

            # Calculate the projection operators for the qubits being measured
            for key, state in qubit_states.items():
                sub_state = "".join([key[index] for index in qubit_indices])
                measurement_operators[sub_state] += state

            for result, operator in measurement_operators.items():
                probability = (operator * final_state).tr()
                self.measurement_statistics.append(
                    MeasurementStatistics(
                        state=(
                            operator * final_state * operator.dag() / probability
                            if probability.real > 0
                            else empty_state
                        ),
                        level=self.level + 1,
                        probability=probability if probability.real > 0 else 0,
                        qubits_measured=qubit_indices,
                        result=result,
                    )
                )

        elif control_type == ControlType.RESET:
            for idx in qubit_indices:
                final_state = (
                    reset_operators[0][idx] * final_state * reset_operators[0][idx].dag()
                    + reset_operators[1][idx] * final_state * reset_operators[1][idx].dag()
                )

            self.measurement_statistics.append(
                MeasurementStatistics(
                    state=final_state,
                    level=self.level + 1,
                    probability=1,
                    qubits_measured=qubit_indices,
                    result=ControlType.RESET,
                )
            )

        for branch in self.measurement_statistics:
            branch_number = branch.build_tree(
                H0,
                c_ops,
                simulation_sections,
                qubit_states,
                reset_operators,
                options,
                branch_number,
            )

        return branch_number

    def extract_outcome_probabilities(self):
        def add_result(result_dict=None):
            if result_dict is None:
                result_dict = {}

            for i, qubit in enumerate(self.qubits_measured):
                result_list = result_dict.setdefault(qubit, [])
                result_list.insert(
                    0,
                    (
                        self.result
                        if self.result == ControlType.RESET
                        else int(self.result[i])
                    ),
                )

            return result_dict

        section = self.level - 1
        outcome_probabilities = {}
        if self.branch_number is not None:
            outcome_probabilities[self.branch_number] = [self.probability, add_result()]
        else:
            for measurement_statistic in self.measurement_statistics:
                outcomes_dict = measurement_statistic.extract_outcome_probabilities()
                if section > -1:
                    for branch_number, outcome in outcomes_dict.items():
                        outcome[0] *= self.probability
                        add_result(outcome[1])
                        outcome_probabilities[branch_number] = outcome
                else:
                    outcome_probabilities.update(outcomes_dict)

        return outcome_probabilities

    def get_branch_nodes(self):
        branch_nodes = {}
        if self.branch_number is not None:
            branch_nodes[self.branch_number] = []
        else:
            for i, measurement_statistic in enumerate(self.measurement_statistics):
                branch_dict = measurement_statistic.get_branch_nodes()
                for branch_number, nodes in branch_dict.items():
                    nodes.insert(0, i)
                    branch_nodes[branch_number] = nodes

        return branch_nodes

    def extract_branch_trajectory(
        self, operators: List["Qobj"], branch: int, step: int = 1, plot_end=False
    ):
        section_dynamics = [[] for i in range(len(operators))]
        sim_t = [[] for i in range(len(operators))]
        dynamics = self.dynamics_results
        measurement_statistic = self
        with log_duration("Branch trajectory extracted in {} seconds."):
            for i in self.get_branch_nodes()[branch]:
                for j in range(len(operators)):
                    section_dynamics[j].extend(
                        [
                            (operators[j] * dynamics[s]).tr()
                            for s in range(0, len(dynamics), step)
                        ]
                    )
                    if measurement_statistic.sim_t is not None:
                        sim_t[j].extend(measurement_statistic.sim_t[::step])
                    else:
                        # If the step doesn't have a time duration due to being a
                        # measurement at time zero or a reset then duplicate the
                        # previous time
                        sim_t[j].extend([0] if len(sim_t[j]) == 0 else [sim_t[j][-1]])
                measurement_statistic = measurement_statistic.measurement_statistics[i]
                if measurement_statistic.branch_number is None:
                    dynamics = measurement_statistic.dynamics_results
                elif plot_end:
                    for j in range(len(operators)):
                        # Append result of final measurement
                        section_dynamics[j].append(
                            (operators[j] * measurement_statistic.state).tr()
                        )
                        sim_t[j].append(sim_t[j][-1])

        return section_dynamics, sim_t


class RTCSQubit(Qubit):
    """Subclass to allow simulation-specific modelling of a qubit."""

    def __init__(
        self,
        index: int,
        resonator: Resonator,
        physical_channel: PhysicalChannel,
        frequency,
        *args,
        N=3,
        anharmonicity=None,
        decay_rate=0.0,
        absorb_rate=0.0,
        dephase_rate=0.0,
        rotating_frame_frequency=0.0,
        **kwargs,
    ):
        super().__init__(index, resonator, physical_channel, *args, **kwargs)
        self._N = 2
        self._rotating_frame_frequency = 0

        self.N = N  # number of qubit energy levels to be simulated
        self.frequency = frequency
        self.rotating_frame_frequency = rotating_frame_frequency
        self.anharmonicity = anharmonicity
        self.decay_rate = decay_rate
        self.absorb_rate = absorb_rate
        self.dephase_rate = dephase_rate
        self._set_qutip_operators()

    def _set_qutip_operators(self):
        # Qutip values / operators
        self.I = qeye(self.N)  # identity
        self.a = destroy(self.N)  # destruction
        self.qutip_N = create(self.N) * destroy(self.N)  # population number
        # Use Pauli if 2 energy levels
        if self.N == 2:
            self.Z = self.I - 2 * self.qutip_N
            self.H0 = -np.pi * self.frequency * self.Z
        else:
            # unitary evolution
            self.H0 = (
                2.0
                * np.pi
                * (self.frequency + 0.5 * (self.qutip_N - 1) * self.anharmonicity)
                * self.qutip_N
            )
        self.cops = [
            self.decay_rate**0.5 * self.a,
            self.absorb_rate**0.5 * self.a.dag(),
            self.dephase_rate**0.5 * self.qutip_N,
        ]  # collapse operators
        self.rho0 = basis(self.N, 0) * basis(self.N, 0).dag()  # Initial state

    @property
    def N(self):
        return self._N

    @property
    def rotating_frame_frequency(self):
        return self._rotating_frame_frequency

    @N.setter
    def N(self, value):
        if self.rotating_frame_frequency != 0 and value != 2:
            raise NotImplementedError(
                "Qubits with a non-zero rotating frame frequency are only implemented "
                "for qubits with 2 energy levels."
            )
        if value < 2:
            raise ValueError("N must be greater than 1")
        self._N = value

    @rotating_frame_frequency.setter
    def rotating_frame_frequency(self, value):
        if self.N != 2 and value != 0:
            raise NotImplementedError(
                "Qubits with more than 2 energy levels are only implemented for a zero "
                "rotating frame frequency"
            )
        self._rotating_frame_frequency = value

    def __getstate__(self):
        dictionary_basis = super().__getstate__()

        del dictionary_basis["I"]
        del dictionary_basis["a"]
        del dictionary_basis["qutip_N"]
        del dictionary_basis["Z"]
        del dictionary_basis["H0"]
        del dictionary_basis["cops"]
        del dictionary_basis["rho0"]

        return dictionary_basis

    def __setstate__(self, state):
        for key, value in state.items():
            self.__dict__[key] = value
        self._set_qutip_operators()


class RTCSResonator(Resonator):
    def __init__(
        self,
        index,
        physical_channel: PhysicalChannel,
        frequency,
        width,
        id_=None,
        *args,
        **kwargs,
    ):
        super().__init__(id_ or f"R{index}", physical_channel, *args, **kwargs)
        self.index = index
        self.frequency = frequency
        # width of lorentzian amplitude about resonator resonant frequency
        self.width = width


class CouplingType(Enum):
    CrossKerr = auto()  # Coupling between a resonator and a qubit
    Exchange = auto()  # Coupling between 2 qubits

    def __repr__(self):
        return self.name


class RTCSCoupling(QuantumComponent, Calibratable):
    """
    Resonators/Qubits are coupled in a very particular way in this model. This object
    holds information about those couplings.
    """

    def __init__(
        self,
        id_: str,
        device_one: QuantumComponent,
        device_two: QuantumComponent,
        frequency,
        coupling_type: CouplingType,
        *args,
        **kwargs,
    ):
        super().__init__(id_, *args, **kwargs)

        self.device_one = device_one
        self.device_two = device_two
        self.coupling_type: CouplingType = coupling_type
        self.frequency = frequency

    def targets(self):
        return [self.device_one, self.device_two]


def _get_highest_id(id_list, prefix, start_id=0):
    highest_id = start_id
    for id in id_list:
        if id.startswith(prefix):
            try:
                highest_id = (
                    int(id[len(prefix) :])
                    if int(id[len(prefix) :]) > highest_id
                    else highest_id
                )
            except:
                pass

    return highest_id


def add_qubit_stack(hw, frequency: float, anharmonicity: float, N: int):
    highest_channel_id = _get_highest_id(list(hw.physical_channels.keys()), "CH")
    highest_baseband_id = _get_highest_id(list(hw.basebands.keys()), "L")

    qubit_channel = f"CH{highest_channel_id+1}"
    qubit_baseband = f"L{highest_baseband_id+1}"
    resonator_channel = f"CH{highest_channel_id+2}"
    resonator_baseband = f"L{highest_baseband_id+2}"
    bb1 = PhysicalBaseband(qubit_baseband, frequency)
    bb2 = PhysicalBaseband(resonator_baseband, 8.0e9)
    hw.add_physical_baseband(bb1, bb2)
    ch1 = PhysicalChannel(qubit_channel, 0.5e-9, bb1, 1)
    ch2 = PhysicalChannel(resonator_channel, 1.0e-9, bb2, 1)
    hw.add_physical_channel(ch1, ch2)

    highest_resonator_id = _get_highest_id(list(hw.quantum_devices.keys()), "R", -1)
    # Currently the frequency of the resonator isn't used in the simulations so we can
    # have the same frequency for all of them without issue
    r = RTCSResonator(highest_resonator_id + 1, ch2, 8.0e9, 1.0e6)
    r.create_pulse_channel(ChannelType.measure, frequency=8.0e9)
    r.create_pulse_channel(ChannelType.acquire, frequency=8.0e9)

    highest_qubit_id = _get_highest_id(list(hw.quantum_devices.keys()), "Q", -1)
    q = RTCSQubit(
        highest_qubit_id + 1,
        r,
        ch1,
        frequency=frequency,
        anharmonicity=anharmonicity,
        N=N,
    )
    q.create_pulse_channel(ChannelType.drive, frequency=frequency)
    q.create_pulse_channel(ChannelType.second_state, frequency=frequency + anharmonicity)

    q_r_coupling = RTCSCoupling(
        f"R{highest_resonator_id+1}<->Q{highest_qubit_id+1}",
        r,
        q,
        10.0e6,
        CouplingType.CrossKerr,
    )

    hw.add_quantum_device(q, r)
    hw.add_couplings(q_r_coupling)

    return q


def add_qubit_coupling(hw, qubit1, qubit2, frequency):
    couple_id = f"{qubit1.id}<->{qubit2.id}"
    q1_q2_coupling = RTCSCoupling(
        couple_id, qubit1, qubit2, frequency, CouplingType.Exchange
    )
    hw.add_couplings(q1_q2_coupling)

    qubit2.add_coupled_qubit(qubit1)
    qubit1.add_coupled_qubit(qubit2)

    q1drive = qubit1.get_pulse_channel(ChannelType.drive)
    q2drive = qubit2.get_pulse_channel(ChannelType.drive)

    qubit1.create_pulse_channel(
        auxiliary_devices=[qubit2],
        channel_type=ChannelType.cross_resonance,
        frequency=q2drive.frequency,
    )

    qubit2.create_pulse_channel(
        auxiliary_devices=[qubit1],
        channel_type=ChannelType.cross_resonance,
        frequency=q1drive.frequency,
    )

    qubit1.create_pulse_channel(
        auxiliary_devices=[qubit2],
        channel_type=ChannelType.cross_resonance_cancellation,
        frequency=q1drive.frequency,
    )

    qubit2.create_pulse_channel(
        auxiliary_devices=[qubit1],
        channel_type=ChannelType.cross_resonance_cancellation,
        frequency=q2drive.frequency,
    )


def apply_setup_to_hardware(hw, rotating_frame=True):
    """Apply the default real-time chip sim hardware setup to the passed-in hardware."""
    bb1 = PhysicalBaseband("L1", 5.0e9)
    bb2 = PhysicalBaseband("L2", 8.0e9)
    bb3 = PhysicalBaseband("L3", 5.09e9)
    bb4 = PhysicalBaseband("L4", 8.5e9)
    hw.add_physical_baseband(bb1, bb2, bb3, bb4)

    ch1 = PhysicalChannel("CH1", 0.5e-9, bb1, 1)
    ch2 = PhysicalChannel("CH2", 1.0e-9, bb2, 1, acquire_allowed=True)
    ch3 = PhysicalChannel("CH3", 0.5e-9, bb3, 1)
    ch4 = PhysicalChannel("CH4", 1.0e-9, bb4, 1, acquire_allowed=True)
    hw.add_physical_channel(ch1, ch2, ch3, ch4)

    r0 = RTCSResonator(0, ch2, 8.0e9, 1.0e6)
    r0.create_pulse_channel(ChannelType.measure, frequency=8.0e9)
    r0.create_pulse_channel(ChannelType.acquire, frequency=8.0e9)

    q0_freq = 5.0e9
    q0 = RTCSQubit(
        0,
        r0,
        ch1,
        q0_freq,
        anharmonicity=-250.0e6,
        N=2 if rotating_frame else 3,
        rotating_frame_frequency=q0_freq if rotating_frame else 0,
    )
    q0.create_pulse_channel(ChannelType.drive, frequency=5.0e9, scale=1.0 + 0.0j)
    q0.create_pulse_channel(ChannelType.second_state, frequency=5.00e9 - 250.0e6)
    q0.mean_z_map_args = [2.136329226009086, -1.0254571168804927]
    q0.pulse_hw_x_pi_2["amp"] = 4368000.007595086

    r1 = RTCSResonator(1, ch4, 8.5e9, 1.0e6)
    r1.create_pulse_channel(ChannelType.measure, frequency=8.5e9)
    r1.create_pulse_channel(ChannelType.acquire, frequency=8.5e9)

    q1_freq = 5.09e9
    q1 = RTCSQubit(
        1,
        r1,
        ch3,
        q1_freq,
        anharmonicity=-250.0e6,
        N=2 if rotating_frame else 3,
        rotating_frame_frequency=q1_freq if rotating_frame else 0,
    )
    q1.create_pulse_channel(ChannelType.drive, frequency=5.09e9, scale=1.0 + 0.0j)
    q1.create_pulse_channel(ChannelType.second_state, frequency=5.09e9 - 250e6)
    q1.mean_z_map_args = [2.1330969344627744, -1.0223925761549326]
    q1.pulse_hw_x_pi_2["amp"] = 4368197.314536925

    q1.add_coupled_qubit(q0)
    q0.add_coupled_qubit(q1)
    q0_r0_coupling = RTCSCoupling("r0<->q0", r0, q0, 10.0e6, CouplingType.CrossKerr)
    q0_q1_coupling = RTCSCoupling("q0<->q1", q0, q1, 4.5e6, CouplingType.Exchange)
    q1_r1_coupling = RTCSCoupling("q1<->r1", q1, r1, 10.0e6, CouplingType.CrossKerr)
    hw.add_couplings(q0_r0_coupling, q0_q1_coupling, q1_r1_coupling)

    q1drive = q1.get_pulse_channel(ChannelType.drive)
    q0drive = q0.get_pulse_channel(ChannelType.drive)

    q0.create_pulse_channel(
        auxiliary_devices=[q1],
        channel_type=ChannelType.cross_resonance,
        frequency=q1drive.frequency,
        scale=20.43888303393216 + 1.9018783700527548j,
    )

    q1.create_pulse_channel(
        auxiliary_devices=[q0],
        channel_type=ChannelType.cross_resonance,
        frequency=q0drive.frequency,
        scale=-19.960910767291143 + 1.7755801175916415j,
    )

    q1.create_pulse_channel(
        auxiliary_devices=[q0],
        channel_type=ChannelType.cross_resonance_cancellation,
        frequency=q1drive.frequency,
        scale=-0.0006383792131791451 - 0.0018484650690712681j,
    )

    q0.create_pulse_channel(
        auxiliary_devices=[q1],
        channel_type=ChannelType.cross_resonance_cancellation,
        frequency=q0drive.frequency,
        scale=-0.002608579454171639 + 0.006639464852534781j,
    )

    hw.add_quantum_device(q0, r0, q1, r1)

    # Simulators don't need automatic calibration.
    hw.is_calibrated = True
    return hw


# noinspection PyPep8Naming
def get_default_RTCS_hardware(repeats=1000, rotating_frame=True):
    model = apply_setup_to_hardware(
        RealtimeSimHardwareModel(repeat_count=repeats), rotating_frame=rotating_frame
    )
    return model


class RealtimeSimHardwareModel(QuantumHardwareModel):
    couplings: List[RTCSCoupling] = []

    def create_engine(self):
        return RealtimeChipSimEngine(self)

    def add_couplings(self, *args):
        self.couplings.extend(list(args))


class RealtimeChipSimEngine(QuantumExecutionEngine):
    """
    Simulation that is built to be as close to our hardware as possible to allow for
    developers, quantum engineers and the fabrication team to be able to run simulations
    and be confident that the results are close to reality.
    """

    model: RealtimeSimHardwareModel

    def __init__(self, model=None, auto_plot=False, sim_qubit_dt=0.25e-10):
        super().__init__(model)
        if not qutip_available:
            raise RuntimeError("qutip unavailable")

        self.auto_plot = auto_plot
        self.sim_qubit_dt = sim_qubit_dt
        self.measurement_statistics: Optional[MeasurementStatistics] = None
        self.sim_t: Optional[List[int]] = None
        self.channel_pulse_data: Optional[dict] = None

    def process_reset(self, position: PositionData):
        """
        When the superclass process_reset is implemented, it should remain empty for the
        simulator.
        """
        pass

    def build_simulator_resets(self, position_map: Dict[str, List[PositionData]]):
        """
        Qubit resets are handled in a unqiue way for the simulator so require their own
        function.
        """
        resets = {}
        for pulse_channel_id, positions in position_map.items():
            for pos in positions:
                if isinstance(pos.instruction, Reset):
                    reset_indices = resets.setdefault(pulse_channel_id, [])
                    reset_indices.append(pos.start)

        return resets

    def optimize(self, instructions: List[Instruction]):
        with log_duration("Instructions optimized in {} seconds."):
            instructions = super().optimize(instructions)
            if not any(inst for inst in instructions if isinstance(inst, Repeat)):
                instructions.append(Repeat(self.model.default_repeat_count))

            for instruction in instructions:
                if isinstance(instruction, PostProcessing):
                    acq = instruction.acquire
                    if acq.mode == AcquireMode.INTEGRATOR:
                        acq.mode = self.model.default_acquire_mode

            return instructions

    def _execute_on_hardware(
        self,
        sweep_iterator: SweepIterator,
        package: QatFile,
        interrupt: Interrupt = NullInterrupt(),
    ) -> Dict[str, np.ndarray]:
        """
        Derivation of the mathematics behind this simulation can be found in the docs
        folder, "Realtime chip simulator mathematical derivation.pdf". Emulate the
        effects of the firmware and quantum hardware for a given input. Before
        instructions are passed to execute they are optimised for this specific function
        so that all measurements occur consecutively and never at the same time across
        all channels.

        Before a measurement is made all quantum channels executing instructions up to
        the measurement must have finished operating and only one one measurement
        channel executes at time. This is achieved by synchronising a measurement
        channel with qubit channels before measurement. This may not be how the hardware
        will handle measurements as it is an inefficent use of time but it makes
        splicing up the simulation into sections between measurements easier for the
        emulator.
        """
        results = {}
        while not sweep_iterator.is_finished():
            sweep_iterator.do_sweep(package.instructions)
            logger.debug(f"Starting sweep #{sweep_iterator.accumulated_sweep_iteration}")

            metadata = {"sweep_iteration": sweep_iterator.get_current_sweep_iteration()}
            interrupt.if_triggered(metadata, throw=True)

            position_map = self.create_duration_timeline(package.instructions)
            pulse_channel_buffers = self.build_pulse_channel_buffers(position_map, True)
            resets = self.build_simulator_resets(position_map)
            buffers = self.build_physical_channel_buffers(pulse_channel_buffers)
            aq_map = self.build_acquire_list(position_map)

            repeats = package.repeat.repeat_count

            sim_length = max(
                [
                    pc.sample_time * len(buffers[pc.id])
                    for pc in self.model.physical_channels.values()
                ]
            )

            # Get simulation time
            self.sim_t = np.arange(0.0, sim_length, self.sim_qubit_dt)

            response_buffers = {}
            resonator_buffers = {}
            for resonator in self.model.resonators:
                measure_channel = resonator.get_measure_channel()
                resonator_buffers[measure_channel.physical_channel_id] = buffers[
                    measure_channel.physical_channel_id
                ]
                response_buffers[measure_channel.physical_channel_id] = np.zeros(
                    shape=[
                        int(repeats),
                        *buffers[measure_channel.physical_channel_id].shape,
                    ],
                    dtype=buffers[measure_channel.physical_channel_id].dtype,
                )

            if all(len(buf) == 0 for buf in resonator_buffers.values()):
                raise NotImplementedError(
                    "There must be at least one qubit measurement to perform a "
                    "simulation"
                )

            # Get resonator channel time steps
            min_resonator_dt = 1.0
            for buffer_idx in resonator_buffers.keys():
                physical_channel = self.model.get_physical_channel(buffer_idx)
                resonator_dt = physical_channel.sample_time
                if min_resonator_dt > resonator_dt:
                    min_resonator_dt = resonator_dt

            # Conversion between resonator and qubit channel time steps
            # TODO: replace with something more stable (interpolation via cubicspline?)
            step = int(min_resonator_dt / self.sim_qubit_dt + 0.5)

            # Convert qubit indices to resonator indices for consistency and change
            # pulse channel to physical channel
            for key in list(resets.keys()):
                item = resets.pop(key)
                pulse_channel = self.model.get_pulse_channel_from_id(key)
                qubit, _ = self.model._resolve_qb_pulse_channel(pulse_channel)
                measure_channel = qubit.get_measure_channel()
                resets[measure_channel.id] = [
                    int(reset * pulse_channel.sample_time / measure_channel.sample_time)
                    for reset in item
                ]

            # Split resonator buffers into segments, one for each measurement channel
            buffer_segments = get_resonator_response_segments(resonator_buffers, resets)

            # find resonator qubit couplings
            target_devices = {}
            lorentzian_responses = {}
            qubit_states = {}
            qutip_size = len(self.model.qubits)
            for buffer_idx, buffer_segment in buffer_segments.items():
                target_device = next(
                    (
                        qb
                        for qb in self.model.qubits
                        if qb.get_measure_channel().physical_channel_id == buffer_idx
                    ),
                    None,
                )
                if target_device is None:
                    continue

                resonator: RTCSResonator = target_device.measure_device
                res_qubit_coupling = next(
                    (
                        coup
                        for coup in self.model.couplings
                        if target_device in coup.targets() and resonator in coup.targets()
                    ),
                    None,
                )
                if res_qubit_coupling is None:
                    continue

                phys_channel = self.model.get_physical_channel(buffer_idx)

                target_devices[buffer_idx] = target_device
                lo = phys_channel.baseband.frequency
                lorentzian_responses[buffer_idx] = {}
                # Generate expected resonator response for qubit in ground and excited
                # state.
                for i, segment in enumerate(buffer_segment):
                    if segment[0] == ControlType.MEASURE:
                        lorentzian_responses[buffer_idx][i] = (
                            get_resonator_response_signal_segment(
                                resonator_input=resonator_buffers[buffer_idx][
                                    segment[1] : segment[2]
                                ],
                                resonator_dt=phys_channel.sample_time,
                                width=resonator.width,
                                res_freq=resonator.frequency,
                                shift=res_qubit_coupling.frequency,
                                lo_freq=lo,
                            )
                        )

            # Generate projection operators for each possible system state
            for out_state in range(2**qutip_size):
                tensor_ops = {}
                output_state_str = format(out_state, f"0{qutip_size}b")
                for idx, device in enumerate(self.model.qubits):
                    if output_state_str[idx] == "1":
                        tensor_ops[idx] = basis(device.N, 1)
                    else:
                        tensor_ops[idx] = basis(device.N, 0)
                qubit_states[output_state_str] = (
                    self.get_tensor(tensor_ops) * self.get_tensor(tensor_ops).dag()
                )

            # Generate reset operators for each qubit
            reset_operators = {
                j: [0 for _ in range(len(self.model.qubits))] for j in range(2)
            }
            for idx, device in enumerate(self.model.qubits):
                P0 = self.get_tensor({idx: basis(device.N, 0) * basis(device.N, 0).dag()})
                P1 = self.get_tensor({idx: basis(device.N, 1) * basis(device.N, 1).dag()})
                Px = self.get_tensor(
                    {
                        idx: basis(device.N, 0) * basis(device.N, 1).dag()
                        + basis(device.N, 1) * basis(device.N, 0).dag()
                    }
                )
                reset_operators[0][idx] = P0
                reset_operators[1][idx] = Px * P1

            # Using the segments, split the simulation into sections, one for each
            # measurement. Consecutive measurements of different channels are considered
            # to be multi-qubit measurements. the first return is the simulation
            # sections and the second the channels measured at the end of each section.
            for key in list(buffer_segments.keys()):
                buffer_segments[target_devices[key].index] = buffer_segments.pop(key)
            simulation_sections: Dict[Section] = get_resonator_response_splicing_indices(
                buffer_segments
            )

            # Construct uncoupled Hamiltonian and find collapse operators
            H0 = 0.0 * self.get_tensor()
            c_ops = []
            for qubit in self.model.qubits:
                qubit: RTCSQubit
                H0 += self.get_tensor({qubit.index: qubit.H0}) - self.get_tensor(
                    {qubit.index: qubit.H0}
                ) * (qubit.rotating_frame_frequency / qubit.frequency)
                c_ops += [self.get_tensor({qubit.index: cop}) for cop in qubit.cops]

            # Find qubit couplings.
            rotation_func = lambda f, T: np.exp(2.0j * np.pi * f * T)
            for coupling in self.model.couplings:
                if coupling.coupling_type == CouplingType.Exchange:
                    allI = []
                    for qb in self.model.qubits:
                        if qb == coupling.device_one:
                            allI.append(qb.a)
                        elif qb == coupling.device_two:
                            allI.append(qb.a.dag())
                        else:
                            allI.append(qb.I)
                    c = 2.0 * np.pi * coupling.frequency * tensor(*allI)
                    delta = (
                        coupling.device_one.rotating_frame_frequency
                        - coupling.device_two.rotating_frame_frequency
                    )

                    previous_index = 0
                    for section_num, section in simulation_sections.items():
                        index = section.indices[0] * step
                        if index != previous_index:
                            section_time = self.sim_t[previous_index : index + 1]
                            section.simulation_time = section_time

                            section.drive_hamiltonian.append(
                                [c, rotation_func(-delta, section_time)]
                            )
                            section.drive_hamiltonian.append(
                                [c.dag(), rotation_func(delta, section_time)]
                            )

                        previous_index = index

            # Upconvert the buffers and use them to create the time dependent
            # Hamiltonian operators
            self.channel_pulse_data = {}
            for device in self.model.quantum_devices.values():
                # Make sure buffers are the same length
                if isinstance(device, Qubit):
                    device: RTCSQubit
                    channel = device.get_drive_channel().physical_channel
                    baseband = channel.baseband
                    buf = buffers[channel.full_id()]
                    dt = channel.sample_time
                    if len(buf) < 2:
                        continue

                    d = spline_time(dt, self.sim_t, buf)
                    self.channel_pulse_data[channel] = d

                    bb = self.model.basebands[baseband.full_id()]
                    LO = bb.frequency
                    d = np.pi * np.exp(UPCONVERT_SIGN * 2.0j * np.pi * LO * self.sim_t) * d

                    # Split the Hamiltonian into measurement sections, specifying which
                    # qubits are to be measured at the end of each section
                    previous_index = 0
                    for section_num, section in simulation_sections.items():
                        index = section.indices[0] * step
                        if index != previous_index:
                            section_time = self.sim_t[previous_index : index + 1]
                            section_buffer = d[previous_index : index + 1]
                            section.simulation_time = section_time

                            section.drive_hamiltonian.append(
                                [
                                    self.get_tensor({device.index: device.a}),
                                    rotation_func(
                                        -device.rotating_frame_frequency, section_time
                                    )
                                    * section_buffer,
                                ]
                            )
                            section.drive_hamiltonian.append(
                                [
                                    self.get_tensor({device.index: device.a.dag()}),
                                    rotation_func(
                                        device.rotating_frame_frequency, section_time
                                    )
                                    * np.conjugate(section_buffer),
                                ]
                            )

                        previous_index = index
                else:
                    channel = device.get_default_pulse_channel().physical_channel
                    dt = self.model.physical_channels[channel.full_id()].sample_time
                    buf = resonator_buffers[channel.full_id()]
                    if len(buf) < 2:
                        continue
                    self.channel_pulse_data[channel.full_id()] = spline_time(
                        dt, self.sim_t, buf
                    )

            options = {"max_step": 1e-11}

            # Initial state.
            rho0 = tensor(*[qb.rho0 for qb in self.model.qubits])

            # Simulate the circuit
            self.measurement_statistics = MeasurementStatistics(rho0)

            # Create a branching structure of all possible outcomes then extract the
            # results of each branch and the probability of it occurring
            self.measurement_statistics.build_tree(
                H0, c_ops, simulation_sections, qubit_states, reset_operators, options
            )
            outcome_probabilities = (
                self.measurement_statistics.extract_outcome_probabilities()
            )

            logger.info("Sim complete.")
            cumulative_probabilities = {}
            counter = 0

            # Cumulative probabilities do not add to 1 as leakage states aren't
            # considered
            for branch, probabilities in outcome_probabilities.items():
                if probabilities[0] != 0:
                    counter += probabilities[0]
                    cumulative_probabilities[counter] = branch

            get_default_logger().debug(str(outcome_probabilities))

            # Generate a random number to choose a possible outcome branch
            for i in range(int(repeats)):
                success = False

                # If a leakage state is chosen try again
                while not success:
                    # Measure qubit state to determine resonator response
                    prob = random()
                    branch = next(
                        (
                            branch
                            for val, branch in cumulative_probabilities.items()
                            if isinstance(val, complex)
                            and prob < val.real
                            or not isinstance(val, complex)
                            and prob < val
                        ),
                        None,
                    )

                    # Just keep trying until we get a value.
                    if branch is not None:
                        success = True

                measurement_results = outcome_probabilities[branch][1]
                for qubit, measurement_result in measurement_results.items():
                    buffer_idx = (
                        self.model.get_qubit(qubit)
                        .get_measure_channel()
                        .physical_channel_id
                    )
                    for j, measurement in enumerate(measurement_result):
                        segment = buffer_segments[qubit][j]
                        if measurement != ControlType.RESET:
                            response_buffers[buffer_idx][i][segment[1] : segment[2]] = (
                                lorentzian_responses[buffer_idx][j][measurement]
                            )

            # Perform post processing to generate a single complex number result for
            # each measurement.
            # Map the result onto the classical registers using
            for channel, aqs in aq_map.items():
                for aq in aqs:
                    response = response_buffers[aq.physical_channel.full_id()][
                        :, aq.start : aq.start + aq.samples
                    ]
                    response_axis = get_axis_map(aq.mode, response)
                    for pp in package.get_pp_for_variable(aq.output_variable):
                        response, response_axis = self.run_post_processing(
                            pp, response, response_axis
                        )

                    var_result = results.setdefault(
                        aq.output_variable,
                        np.empty(
                            sweep_iterator.get_results_shape(response.shape),
                            response.dtype,
                        ),
                    )
                    sweep_iterator.insert_result_at_sweep_position(var_result, response)

            if self.auto_plot:
                self.plot_pulses()
                self.plot_dynamics()

        return results

    def plot_pulses(self, channels: List[str] = None):
        """
        Plot pulses used to drive the system. Pulses are shown before being transformed
        by the baseband frequnecy.

        :param channels: List of channel ids to plot. If None then all channels are plotted.
            Defaults to None.
        """
        if channels is None:
            channels = list(self.channel_pulse_data.keys())
        fig, axes = create_subplots(len(channels))
        axes = iter(axes)

        for channel in channels:
            pulse = self.channel_pulse_data[channel]
            ax = next(axes)

            ax.set_title(channel)
            ax.plot(self.sim_t, pulse.real, label="I")
            ax.plot(self.sim_t, pulse.imag, label="Q")
            ax.set_xlabel("Time")
            ax.set_ylabel("Amp")
            ax.legend(shadow=True, fancybox=True)

        fig.suptitle("Pulses", fontsize=16)
        plt.tight_layout()
        plt.show()

    def plot_dynamics(
        self,
        operator_info: List[Union[OperatorInfo, int]] = None,
        branches: List[int] = None,
        step: int = 1,
    ):
        """
        Plot the dynamics of operator expectation values for a simulation run on this
        hardware. By default the pauli operators will be plot for each qubit but custom
        operators can be specified.

        :param operator_info: Use the OperatorInfo object to specify an operator to
            plot, the id of the qubit the operator should act on and the name of the
            operator. Alternatively qubit indices (int) can be specified in which case
            the Pauli x, y and z expectations will be plotted for these qubits. If None
            then the Pauli operators for every qubit will be plotted. Defaults to None.
        :param branches: Specify a list of indices for the branches you wish to plot.
            This is only really relevent if the circuit includes mid-circuit
            measurements, in which case there will be a different trajectory for each
            possible measurement outcome. If None then the first trajectory will be
            plotted which is identical to all other trajectories for circuits without
            mid-circuit measurements. Defaults to None.
        :param step: The step length between expectation calculations along the
            trajectory. 1 gives the highest resolution but takes longer to calculate.
            Defaults to 1.
        """
        if branches is None:
            branches = [0]
            plot_end = False
        else:
            plot_end = True

        # we separate out operators and their names as it is more efficient to process
        # all operators at once
        operators = []
        operator_names = []
        # Plot all qubits if not specified
        if operator_info is None:
            operator_info = list(range(len(self.model.qubits)))
        for i, op in enumerate(operator_info):
            # if an int is given, plot pauli x, y and z for the qubit with the
            # corresponding id
            if isinstance(op, int):
                device = self.model.qubits[op]
                operators.append(
                    self.get_tensor(
                        {
                            op: basis(device.N, 0) * basis(device.N, 1).dag()
                            + basis(device.N, 1) * basis(device.N, 0).dag()
                        }
                    )
                )
                operator_names.append((op, "Pauli X"))

                operators.append(
                    self.get_tensor(
                        {
                            op: 1.0j
                            * (
                                -basis(device.N, 0) * basis(device.N, 1).dag()
                                + basis(device.N, 1) * basis(device.N, 0).dag()
                            )
                        }
                    )
                )
                operator_names.append((op, "Pauli Y"))

                operators.append(
                    self.get_tensor(
                        {
                            op: basis(device.N, 0) * basis(device.N, 0).dag()
                            - basis(device.N, 1) * basis(device.N, 1).dag()
                        }
                    )
                )
                operator_names.append((op, "Pauli Z"))
            elif isinstance(op.qubit_id, int):
                operators.append(self.get_tensor({op.qubit_id: op.operator}))
                operator_names.append((op.qubit_id, op.name))
            else:
                operators.append(op.operator)
                operator_names.append((f"not specified {i}", op.name))

        for branch in branches:
            (
                branch_trajectories,
                branch_time,
            ) = self.measurement_statistics.extract_branch_trajectory(
                operators, branch, step, plot_end
            )
            # Sort by qubit id
            trajectory_dict = {}
            for name, trajectory, time in zip(
                operator_names, branch_trajectories, branch_time
            ):
                item = trajectory_dict.setdefault(name[0], [])
                item.append((name[1], trajectory, time))

            for id, operator_info in trajectory_dict.items():
                fig, axes = create_subplots(len(operator_info))
                fig.suptitle(f"Branch {branch}, Qubit {id}")
                axes = iter(axes)
                for name, trajectory, t in operator_info:
                    ax = next(axes)
                    ax.set_title(name)
                    ax.set_xlabel("Time")
                    ax.set_ylabel("<op>")
                    ax.plot(t, np.array(trajectory).real)
                fig.tight_layout()

        plt.show()

    def get_branches(self):
        """Return the branch ids along with the measurements associated with them.

        :Example:

        ``{branch_id: [branch_probability, {qubit: [resets_and_measurement_outcomes]}]}``

          Where:

          - branch_id: id to call when plotting specific branches
          - branch_probability: probability branch occured
          - qubit: qubit number
          - resets_and_measurement_outcomes: the outcome of measurements for this branch
            in order. Also includes when a reset occured.

        :returns: A dictionary mapping branch ids to their associated measurements.
        :rtype: dict
        """
        return self.measurement_statistics.extract_outcome_probabilities()

    def get_tensor(self, ops: dict = None):
        if ops is None:
            ops = {}
        allI = [qb.I for qb in self.model.qubits]
        for _idx, op in ops.items():
            allI[_idx] = op
        return tensor(*allI)


def spline_time(dt, sim_t, buffer):
    """
    Perform a cubic spline on the buffer to make the number of points the same as the
    number of points in the simulation time
    """
    t = np.linspace(0.0, (len(buffer) - 1) * dt, len(buffer))
    if t[-1] < sim_t[-1]:
        extra_t = np.arange(t[-1] + dt, sim_t[-1], dt)
        extra_d = np.zeros(len(extra_t), dtype=np.complex128)

        t = np.concatenate((t, extra_t))
        buffer = np.concatenate((buffer, extra_d))

    buffer = CubicSpline(t, buffer)(sim_t)

    return buffer


def tukey_window(N, alpha=0.1):
    n = np.linspace(0.0, 1.0, N)
    x = 0.5 * (1 - np.cos(2.0 * np.pi * n / alpha))
    start = int(N * alpha * 0.5)
    x[start:].fill(1.0)
    y = x + np.flip(x, axis=0) - 1.0
    return y


def get_resonator_response_segments(buffers, resets):
    zero_tolerance = 1e-6
    zero_tolerance_sq = zero_tolerance**2

    max_len = max([len(buffer) for buffer in buffers.values()])
    buffer_inside = {k: False for k, v in buffers.items()}
    buffer_segments = {}
    i = 0
    while i < max_len:
        for buffer_idx, buffer in buffers.items():
            reset_indices = resets.get(buffer_idx, [])
            if i < len(buffer):
                if buffer[i] ** 2 > zero_tolerance_sq:
                    if not buffer_inside[buffer_idx]:
                        segment = buffer_segments.setdefault(buffer_idx, [])
                        segment.append([ControlType.MEASURE, i])
                        buffer_inside[buffer_idx] = True
                elif buffer_inside[buffer_idx]:
                    buffer_segments[buffer_idx][-1].append(i)
                    buffer_inside[buffer_idx] = False
                if i in reset_indices:
                    segment = buffer_segments.setdefault(buffer_idx, [])

                    # If currently constructing a measurement section then reset
                    # must be before this measurement
                    if buffer_inside[buffer_idx]:
                        segment.insert(-1, [ControlType.RESET, i, i])
                    else:
                        segment.append([ControlType.RESET, i, i])
            elif buffer_inside[buffer_idx] == True:
                buffer_segments[buffer_idx][-1].append(i)
                buffer_inside[buffer_idx] = False
        i += 1
    for buffer_idx, buffer_segment in buffer_segments.items():
        if buffer_inside[buffer_idx]:
            buffer_segment[-1].append(i - 1)

    return buffer_segments


def get_resonator_response_splicing_indices(buffer_segments):
    """
    Take the buffer segments and parse them to find where the qubit buffer needs to be
    split in order to perform mid-circuit measurements.
    """
    previous_start_indice = 0
    section_num = -1
    simulation_sections: Dict[Section] = {}
    ordered_segments = []
    for qubit, segments in buffer_segments.items():  # Order segments from first to last
        for segment in segments:
            ordered_segments.append([qubit, *segment])
    # sort by last index if first indices are the same to make sure reset are before
    # measures
    ordered_segments.sort(key=lambda x: (x[2], x[3]))

    previous_control_type = ordered_segments[0][1]
    for segment in ordered_segments:
        indices = [segment[2], segment[3]]
        if indices[0] == previous_start_indice and segment[1] == previous_control_type:
            if section_num == -1:  # Catch case measurement occurs at time 0
                section_num += 1
                simulation_sections[section_num] = Section(
                    indices, [segment[0]], segment[1]
                )
            else:
                simulation_sections[section_num].qubits.append(segment[0])
        else:
            section_num += 1
            simulation_sections[section_num] = Section(indices, [segment[0]], segment[1])

        previous_start_indice = indices[0]
        previous_control_type = segment[1]

    return simulation_sections


def get_resonator_response_signal_segment(
    resonator_input, resonator_dt, width, res_freq, shift, lo_freq
):
    # Peak of resonator resonant frequency relative to LO freq
    centre0 = res_freq - lo_freq
    # Peak of resonator resonant frequency relative to LO freq with excited qubit
    centre1 = centre0 + shift

    sig0 = tukey_window(len(resonator_input)) * resonator_input

    o = np.fft.fft(sig0)
    f = np.fft.fftfreq(len(o), d=resonator_dt)

    def lorentzian(u, centre, width):
        return 1 / (1 + ((u - centre) / (width / 2)) ** 2)

    # Create Lorentzian curves in the Fourier transform of the buffer segment
    # representing the resonator spectrum
    resonator_output_0 = np.fft.ifft(
        o * lorentzian(f, centre0, width)
    )  # Response if qubit in ground state
    resonator_output_1 = np.fft.ifft(
        o * lorentzian(f, centre1, width)
    )  # Response if qubit in excited state

    return [resonator_output_0, resonator_output_1]


def get_simple_resonator_response(
    qubit_dt, qubit_z, resonator_dt, resonator_iq, width, res_freq, shift, lo_freq
):
    centre0 = res_freq - lo_freq
    centre1 = centre0 + shift

    iq_t = np.linspace(
        0.0, resonator_dt * len(resonator_iq), len(resonator_iq), endpoint=False
    )  # Create array of sample times
    resonator_iq *= np.exp(-UPCONVERT_SIGN * 2.0j * np.pi * (res_freq - lo_freq) * iq_t)

    ZERO_TOLERANCE = 1e-6

    abssq = np.abs(resonator_iq) ** 2
    ztsq = ZERO_TOLERANCE**2

    inside = False
    start = 0
    i = 0
    segments = []
    while i < len(abssq):
        if inside:
            if abssq[i] < ztsq:
                inside = False
                segments.append((start, i))
        else:
            if abssq[i] > ztsq:
                inside = True
                start = i
        i += 1
    if inside:
        segments.append((start, i))

    simulate_continuous_measurement = (
        len(segments) == 1
        and segments[0][0] == 0
        and segments[0][1] == len(resonator_iq) - 1
    )

    # segments is now a list of start and stop index of non-zero segments for the
    # resonator response

    # for now assume qubit_t and resonator_t are linearly spaced arrays starting at zero

    # qubit_dt = qubit_t[1] - qubit_t[0]
    # resonator_dt = resonator_t[1] - resonator_t[0]

    factor = qubit_dt / resonator_dt

    qsegments = [(int(round(factor * s[0])), int(round(factor * s[1]))) for s in segments]

    qzsegments = []
    for start, stop in qsegments:
        z = np.mean(qubit_z[start:stop])
        if np.isnan(z):
            z = qubit_z[len(qubit_z) - 1]
        qzsegments.append(z)

    input = resonator_iq

    output = np.zeros(shape=input.shape, dtype=input.dtype)

    def flat_top_window(N):
        n = np.linspace(0.0, 1.0, N)
        a0 = 0.35875
        a1 = 0.48829
        a2 = 0.14128
        a3 = 0.01168
        return (
            a0
            - a1 * np.cos(2.0 * np.pi * n)
            + a2 * np.cos(4.0 * np.pi * n)
            - a3 * np.cos(6.0 * np.pi * n)
        )

    for s, qz in zip(segments, qzsegments):
        p0 = 0.5 * (qz + 1)
        p1 = 1 - p0

        i, j = s
        x = input[i:j]
        # if simulate_continuous_measurement:
        #     sig0 = x # flat_top_window(len(x))*x
        # else:
        sig0 = tukey_window(len(x)) * x

        o = np.fft.fft(sig0)
        f = np.fft.fftfreq(len(o), d=resonator_dt)

        def L(u, centre, width):
            return 1 / (1 + ((u - centre) / (width / 2)) ** 2)

        S = p0 * L(f, centre0, width) + p1 * L(f, centre1, width)
        o *= S
        o = np.fft.ifft(o)

        # plt.subplots()
        # plt.plot(sig0.real)
        # plt.plot(sig0.imag)
        # plt.plot(o.real)
        # plt.plot(o.imag)
        # plt.show()

        # todo: maybe method to force signal continuity should be here, but this method
        #   messes up sometimes
        # diff = o[0] - sig0[0]
        # o -= diff

        output[i:j] = o

        if np.isnan(output[i]):
            raise Exception()

    return output


def create_subplots(plot_num):
    square = int(np.sqrt(plot_num))
    extra_rows = int(np.ceil((plot_num - square**2) / square))
    fig, axes = plt.subplots(nrows=square + extra_rows, ncols=square, squeeze=True)
    if isinstance(axes, Iterable):
        axes = list(axes.flatten())
    else:
        axes = [axes]
    for i in range(len(axes) - plot_num):
        fig.delaxes(axes.pop())

    return fig, axes



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
from compiler_config.config import (
    MetricsType,
    OptimizationConfig,
    Qiskit,
    QiskitOptimizations,
    Tket,
    TketOptimizations,
)
from qiskit import QuantumCircuit, qasm2, transpile
from qiskit.transpiler import TranspilerError

from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.metrics import MetricsMixin
from qat.purr.integrations.tket import run_tket_optimizations
from qat.purr.utils.logger import get_default_logger
from qat.purr.utils.logging_utils import log_duration

log = get_default_logger()


class DefaultOptimizers(MetricsMixin):
    def __init__(self, metrics=None):
        super().__init__()
        self.compilation_metrics = metrics

    def optimize_qasm(
        self,
        qasm_string,
        hardware: QuantumHardwareModel,
        optimizations: OptimizationConfig,
    ):
        """Run all available optimizers on this QASM program."""
        with log_duration("QASM optimization took {} seconds."):
            if (
                isinstance(optimizations, Tket)
                and optimizations.tket_optimizations != TketOptimizations.Empty
            ):
                qasm_string = run_tket_optimizations(
                    qasm_string, optimizations.tket_optimizations, hardware
                )

            # TODO: [QK] Spend time looking at qiskit optimization and seeing if it's
            #   worth keeping around.
            if (
                isinstance(optimizations, Qiskit)
                and optimizations.qiskit_optimizations != QiskitOptimizations.Empty
            ):
                qasm_string = self.run_qiskit_optimization(
                    qasm_string, optimizations.qiskit_optimizations
                )

            self.record_metric(MetricsType.OptimizedCircuit, qasm_string)
            return qasm_string

    def run_qiskit_optimization(self, qasm_string, level):
        """
        TODO: [QK] Current setup is unlikely to provide much benefit, refine settings
            before using.
        """
        if level is not None:
            try:
                optimized_circuits = transpile(
                    QuantumCircuit.from_qasm_str(qasm_string),
                    basis_gates=["u1", "u2", "u3", "cx"],
                    optimization_level=level,
                )
                qasm_string = qasm2.dumps(optimized_circuits)
            except TranspilerError as ex:
                log.warning(f"Qiskit transpile pass failed. {str(ex)}")

        return qasm_string



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
from copy import deepcopy
from numbers import Number
from typing import List

from compiler_config.config import TketOptimizations
from pytket import Bit, Circuit, Qubit
from pytket._tket.architecture import Architecture, RingArch
from pytket._tket.circuit import CustomGateDef
from pytket._tket.predicates import (
    ConnectivityPredicate,
    DirectednessPredicate,
    MaxNQubitsPredicate,
    NoMidMeasurePredicate,
)
from pytket._tket.transform import Transform
from pytket.passes import (
    CliffordSimp,
    ContextSimp,
    DecomposeArbitrarilyControlledGates,
    DecomposeBoxes,
    DefaultMappingPass,
    FullPeepholeOptimise,
    GlobalisePhasedX,
    KAKDecomposition,
    PeepholeOptimise2Q,
    RebaseTket,
    RemoveBarriers,
    RemoveDiscarded,
    RemoveRedundancies,
    SequencePass,
    SimplifyMeasured,
    SynthesiseTket,
    ThreeQubitSquash,
)
from pytket.qasm import circuit_to_qasm_str
from pytket.qasm.qasm import NOPARAM_COMMANDS, PARAM_COMMANDS, QASMUnsupportedError
from qiskit.circuit.library import CXGate, UGate
from sympy import pi, sympify

from qat.purr.compiler.execution import QuantumHardwareModel
from qat.purr.integrations.qasm import BitRegister, Qasm2Parser, QasmContext, QubitRegister
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class TketBuilder:
    """
    Builds a Tket circuit using a builder API. No direct relation to our other builders
    as the API is a little too different.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.circuit = Circuit()

    def qreg(self, name, size):
        self.circuit.add_q_register(name, size)

    def creg(self, name, size):
        self.circuit.add_c_register(name, size)

    def barrier(self, qubits):
        qubits = [qubits] if not isinstance(qubits, List) else qubits
        self.circuit.add_barrier(qubits)

    def measure(self, qubits, bits, conditions=None):
        qubits = [qubits] if not isinstance(qubits, List) else qubits
        bits = [bits] if not isinstance(bits, List) else bits
        conditions = conditions or {}
        for qubit, bit in zip(qubits, bits):
            self.circuit.Measure(qubit, bit, **conditions)

    def is_basic_gate(self, name):
        return self._get_basic_gate_def(name) is not None

    def _get_basic_gate_def(self, name):
        return PARAM_COMMANDS.get(name, None) or NOPARAM_COMMANDS.get(name, None)

    def gate(self, name, qubits, params=None, conditions=None):
        target_gate = self._get_basic_gate_def(name)
        if target_gate is None:
            raise ValueError(f"Can't find a definition for {name}.")

        conditions = conditions or {}
        params = params or []
        qubits = [qubits] if not isinstance(qubits, List) else qubits
        params = [params] if not isinstance(params, List) else params
        self.circuit.add_gate(target_gate, params, qubits, **conditions)

    def custom_gate(self, gate_def: CustomGateDef, qubits, params=None, conditions=None):
        conditions = conditions or {}
        params = params or []
        qubits = [qubits] if not isinstance(qubits, List) else qubits
        params = [params] if not isinstance(params, List) else params
        self.circuit.add_custom_gate(gate_def, params, qubits, **conditions)


class TketQasmParser(Qasm2Parser):
    """
    QASM parser that turns QASM into Tket structures. Switch to Tkets QASM parser when it's
    more mature.
    """

    def process_qreg(self, node, context: QasmContext, builder: TketBuilder, **kwargs):
        if node.name in context.registers.quantum:
            return
        context.registers.quantum[node.name] = QubitRegister(
            [Qubit(node.name, val) for val in range(node.size)]
        )
        builder.qreg(node.name, node.size)

    def process_creg(self, node, context, builder: TketBuilder, **kwargs):
        if node.name in context.registers.classic:
            return
        context.registers.classic[node.name] = BitRegister(
            [Bit(node.name, val) for val in range(node.size)]
        )
        builder.creg(node.name, node.size)

    def process_gate_definition(self, node, context, _, **kwargs):
        # Just a declaration, gates are used when CustomUnitary is used
        context.gates[node.name] = node

    def _get_staggered_qubit_params(self, node, context):
        def transform_to_halfturn(num):
            try:
                return sympify(num) / pi
            except:
                raise ValueError("Cannot parse angle: {}".format(num))

        res = self._expand_to_match_registers(
            self._get_qubits(node, context),
            [
                transform_to_halfturn(param) if isinstance(param, Number) else param
                for param in self._get_parameters(node, context)
            ],
        )

        return res

    def process_gate(
        self,
        method,
        context: QasmContext,
        builder: TketBuilder,
        conditions=None,
        **kwargs,
    ):
        # For ease-of-use we just process CX and U as a normal gate, but the QASM nodes
        # don't have names. Infer them in this case.
        if isinstance(method.op, UGate):
            gate_name = "U"
        elif isinstance(method.op, CXGate):
            gate_name = "CX"
        else:
            gate_name = method.name

        # TODO: Right now we just flatten all gates. Should move to use the
        #   MethodGateDef of Tket, but as that functionality is currently broken we
        #   lose nothing here.
        for qbs, args in self._get_staggered_qubit_params(method, context):
            if builder.is_basic_gate(gate_name):
                builder.gate(gate_name, qbs, args, conditions)
            else:
                super().process_gate(
                    method, context, builder, conditions=conditions, **kwargs
                )

    def process_if(self, node, context, builder: TketBuilder, *kwargs):
        left = self._resolve_variable(node.children[0], context)
        right = self._resolve_value(node.children[1], context)

        if isinstance(left, BitRegister):
            left = left.bits
        else:
            left = [left]

        conditions = {"condition_bits": left, "condition_value": right}
        self.walk_node(node.children[2], context, builder, conditions=conditions)

    def process_reset(self, node, context, builder, conditions=None, **kwargs):
        for qubits, _ in self._get_staggered_qubit_params(node, context):
            builder.gate("reset", qubits, conditions)

    def process_measure(
        self,
        node,
        context: QasmContext,
        builder: TketBuilder,
        conditions=None,
        **kwargs,
    ):
        for qubits, bits in self._get_staggered_qubit_params(node, context):
            builder.measure(qubits, bits, conditions)

    def process_unitary(
        self, node, context, builder: TketBuilder, conditions=None, **kwargs
    ):
        self.process_gate(node, context, builder, conditions, **kwargs)

    def process_cnot(self, node, context, builder, conditions=None, **kwargs):
        self.process_gate(node, context, builder, conditions, **kwargs)

    def process_barrier(self, node, context, builder: TketBuilder, **kwargs):
        barrier_qubits = []
        for qubits, _ in self._get_staggered_qubit_params(node, context):
            if isinstance(qubits, List):
                barrier_qubits.extend(qubits)
            else:
                barrier_qubits.append(qubits)

        builder.barrier(barrier_qubits)

    def process_program(self, builder, qasm):
        self._walk_program(builder, qasm, QasmContext())
        return builder


def fetch_default_passes(architecture, opts, pass_list: List = None, add_delay=True):
    pass_list = pass_list or []
    if TketOptimizations.DefaultMappingPass in opts:
        pass_list.append(DefaultMappingPass(architecture, add_delay))

    return pass_list


def _full_stopalize(e):
    exception = str(e)
    return exception if exception.endswith(".") else f"{exception}."


def apply_default_transforms(circuit, architecture, opts):
    transform_list = []
    if TketOptimizations.DirectionalCXGates in opts:
        transform_list.append(Transform.DecomposeCXDirected(architecture))
        transform_list.append(Transform.RemoveRedundancies())

    for transform in transform_list:
        transform.apply(circuit)


def check_validity(circuit, architecture):
    predicates = [
        (
            "Failed to apply circuit to hardware topology.",
            ConnectivityPredicate(architecture),
        ),
        (
            "Failed to satisfy 2QB gate constraints.",
            DirectednessPredicate(architecture),
        ),
        ("No mid-circuit measurements allowed.", NoMidMeasurePredicate()),
    ]

    failed_preds = [message for message, pred in predicates if not pred.verify(circuit)]
    if any(failed_preds):
        raise ValueError(f"Verification failed: {' '.join(failed_preds)}")


def optimize_circuit(circ, architecture, opts):
    """
    Apply all Tket optimizations to the circuit provided as defined in the options flag.
    """
    try:
        # By default we want to decompose Tket-specific circuits and then rebase into
        # the form that Tket can optimize most efficiently.
        passes = [DecomposeBoxes(), SynthesiseTket()]

        if TketOptimizations.ContextSimp in opts:
            passes.append(ContextSimp(allow_classical=True))

        # TODO: FullPeepholeOptimise fails due to bug in tket. Find out what's wrong and
        #   fix/flag.
        if TketOptimizations.FullPeepholeOptimise in opts:
            passes.append(FullPeepholeOptimise())

        if TketOptimizations.CliffordSimp in opts:
            passes.append(CliffordSimp(allow_swaps=True))

        if TketOptimizations.DecomposeArbitrarilyControlledGates in opts:
            passes.append(DecomposeArbitrarilyControlledGates())

        # if TketOptimizations.EulerAngleReduction in opts:
        #     passes.append(EulerAngleReduction())

        if TketOptimizations.GlobalisePhasedX in opts:
            passes.append(GlobalisePhasedX())

        # if TketOptimizations.GuidedPauliSimp in opts:
        #     passes.append(GuidedPauliSimp())

        if TketOptimizations.KAKDecomposition in opts:
            passes.append(KAKDecomposition())

        # if TketOptimizations.OptimisePhaseGadgets in opts:
        #     passes.append(OptimisePhaseGadgets())
        #
        # if TketOptimizations.PauliSimp in opts:
        #     passes.append(PauliSimp())
        #
        # if TketOptimizations.PauliSquash in opts:
        #     passes.append(PauliSquash())

        if TketOptimizations.PeepholeOptimise2Q in opts:
            passes.append(PeepholeOptimise2Q())

        if TketOptimizations.RemoveDiscarded in opts:
            passes.append(RemoveDiscarded())

        if TketOptimizations.RemoveBarriers in opts:
            passes.append(RemoveBarriers())

        if TketOptimizations.RemoveRedundancies in opts:
            passes.append(RemoveRedundancies())

        if TketOptimizations.ThreeQubitSquash in opts:
            passes.append(ThreeQubitSquash())

        if TketOptimizations.SimplifyMeasured in opts:
            passes.append(SimplifyMeasured())

        # TODO: Use HW model object to create fetch calibration data for optimal
        #   routing.
        passes = fetch_default_passes(architecture, opts, passes)

        passes += [RebaseTket()]

        # Not everything in the list is a pass, we've also got transforms.
        # Everything in the list should have an apply function though.
        # for pass_ in passes:
        SequencePass(passes).apply(circ)
        apply_default_transforms(circ, architecture, opts)
        check_validity(circ, architecture)

        log.info("Tket optimizations complete.")
        return True
    except (ValueError, IndexError, RuntimeError) as e:
        log.warning(
            f"Tket could not compile circuit due do exception: {_full_stopalize(e)}. "
            "Skipping this optimization pass."
        )
        return False


def get_coupling_subgraphs(couplings):
    """
    Given a list of connections which potentially describe a disconnected graph,
    this returns a list of connected subgraphs
    """
    subgraphs = []
    for coupling in couplings:
        subgraph_added_to = None
        for subgraph in subgraphs:
            # check whether coupling joins to subgraph
            if set(coupling) & {qubit for connection in subgraph for qubit in connection}:
                if subgraph_added_to is None:
                    subgraph.append(coupling)  # add the coupling to the subgraph
                    subgraph_added_to = subgraph
                else:
                    subgraph_added_to.extend(
                        subgraph
                    )  # if coupling joins to 2nd subgraph, the subgraphs combine
                    subgraphs.remove(subgraph)
        if (
            subgraph_added_to is None
        ):  # if coupling doesn't join to any of the current subgraphs, make it a new one
            subgraphs.append([coupling])
    return subgraphs


def run_tket_optimizations(qasm_string, opts, hardware: QuantumHardwareModel) -> str:
    """
    Runs tket-based optimizations and modifications. Routing will always happen no
    matter the level.

    Will run optimizations in sections if a full suite fails until a minimal subset of
    passing optimizations is found.
    """

    try:
        tket_builder: TketBuilder = TketQasmParser().parse(TketBuilder(), qasm_string)
        circ = tket_builder.circuit
        log.info(f"Number of gates before tket optimization: {circ.n_gates}")
    except Exception as e:  # Parsing is too fragile, can cause almost any exception.
        log.warning(
            f"Tket failed during QASM parsing with error: {_full_stopalize(e)}. "
            "Skipping this optimization pass."
        )
        return qasm_string

    couplings = deepcopy(hardware.qubit_direction_couplings)
    optimizations_failed = False
    architecture = None
    if any(couplings):
        # Without default remapping pass multi-qubit gates don't get moved around, so
        # trying to apply them to a limited subset of qubits provides no value.
        logical_qubit_map = {q.index: i for i, q in enumerate(hardware.qubits)}
        for coupling in couplings:
            control_ind, target_ind = coupling.direction
            coupling.direction = (
                logical_qubit_map[control_ind],
                logical_qubit_map[target_ind],
            )
        if TketOptimizations.DefaultMappingPass not in opts:
            architecture = Architecture([val.direction for val in couplings])
            optimizations_failed = not optimize_circuit(circ, architecture, opts)
        else:
            coupling_qualities = list({val.quality for val in couplings})
            coupling_qualities.sort(reverse=True)
            for quality_level in coupling_qualities:
                filtered_couplings = [
                    val.direction for val in couplings if val.quality >= quality_level
                ]
                coupling_subgraphs = get_coupling_subgraphs(filtered_couplings)
                for subgraph in coupling_subgraphs:
                    if circ.n_qubits <= len(
                        set([qubit for coupling in subgraph for qubit in coupling])
                    ):
                        architecture = Architecture(subgraph)
                        optimizations_failed = not optimize_circuit(
                            circ, architecture, opts
                        )
                        if not optimizations_failed:
                            break
                    else:
                        optimizations_failed = True
                if not optimizations_failed:
                    break
    else:
        architecture = RingArch(len(hardware.qubits))
        optimizations_failed = not optimize_circuit(circ, architecture, opts)

    # If our optimizations failed but we want the mapping pass, apply that by itself.
    if optimizations_failed:
        if architecture is None:
            raise ValueError(
                "Unable to resolve hardware instance for fall-back optimizations."
            )

        delay_failed = False
        try:
            # DelayMeasure throws on failure, and we want to raise our own errors for
            # this.
            SequencePass(fetch_default_passes(architecture, opts)).apply(circ)
        except RuntimeError:
            delay_failed = True

        # If the delay fails, try with a more limited subset.
        if delay_failed:
            try:
                # Tket just throws an exception if the list is none, so skip if that's
                # the case.
                default_passes = fetch_default_passes(architecture, opts, add_delay=False)
                if len(default_passes) > 0:
                    SequencePass(default_passes).apply(circ)
            except RuntimeError as e:
                message = str(e)
                if MaxNQubitsPredicate.__name__ in message:
                    raise ValueError(
                        f"Circuit uses {len(circ.qubits)} qubits, "
                        f"only {len(architecture.nodes)} available."
                    )

                raise e

        apply_default_transforms(circ, architecture, opts)
        check_validity(circ, architecture)

    try:
        qasm_string = circuit_to_qasm_str(circ)
        log.info(f"Number of gates after tket optimization: {circ.n_gates}")
    except (QASMUnsupportedError, RuntimeError) as e:
        log.warning(
            f"Error generating QASM from Tket circuit: {_full_stopalize(e)}. "
            "Skipping this optimization pass."
        )

    # TODO: Return result object with more information about compilation/errors
    return qasm_string



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
from typing import List, Optional

from compiler_config.config import MetricsType

from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class _FlagFieldValidation(type):
    """
    Validate that the CompilationsMetrics class has type hints for all the metrics,
    which'll be dynamically assigned later.
    """

    def __new__(mcs, name, inheritance, attributes):
        hint_names = set(attributes.get("__annotations__", {}).keys())
        snake_cased_flags = [
            val.snake_case_name()
            for val in list(dict(MetricsType.__members__).values())
            if not val.is_composite()
        ]
        missing_fields = [val for val in snake_cased_flags if val not in hint_names]
        if any(missing_fields):
            raise ValueError(
                "CompilationMetrics is missing type hints for "
                f"{', '.join(missing_fields)}."
            )

        return super().__new__(mcs, name, inheritance, attributes)


class CompilationMetrics(metaclass=_FlagFieldValidation):
    """
    Container object for all metrics generated during the compilation pipeline. All
    fields are generated from the MetricsType flag after the names are snake-cased and
    hold the value associated with that particular flag.
    """

    optimized_circuit: Optional[str]
    optimized_instruction_count: Optional[int]

    def __init__(self, enabled_metrics=None):
        self.enabled_metrics: Optional[MetricsType] = enabled_metrics or MetricsType.Default
        for key in [val.snake_case_name() for val in self._target_metrics()]:
            setattr(self, key, None)

    def _target_metrics(self) -> List[MetricsType]:
        """Get a list of the enum types that we should function on."""
        return [
            val
            for val in list(dict(MetricsType.__members__).values())
            if not val.is_composite()
        ]

    def enable(self, enabled_metrics: MetricsType, overwrite=False):
        """
        Enable these sets of metrics for collection. If overwrite is True then the
        passed-in values will overwrite existing ones.
        """
        if enabled_metrics is None:
            return

        if overwrite:
            self.enabled_metrics = enabled_metrics
        else:
            self.enabled_metrics = self.enabled_metrics | enabled_metrics

    def are_enabled(self, metric: MetricsType):
        return self.enabled_metrics is not None and metric in self.enabled_metrics

    def record_metric(self, metric: MetricsType, value):
        if not self.are_enabled(metric):
            return

        setattr(self, metric.snake_case_name(), value)

    def get_metric(self, metric: MetricsType):
        return getattr(self, metric.snake_case_name())

    def as_dict(self):
        """Generates a dictionary of the valid metrics."""
        return {
            met.snake_case_name(): self.get_metric(met) for met in self._target_metrics()
        }

    def merge(self, other: "CompilationMetrics"):
        if not isinstance(other, type(self)):
            raise TypeError(
                "Argument other must be of type Compilation metrics, not of type: "
                f"{type(other)}"
            )
        for metric in other._target_metrics():
            if self.get_metric(metric) is None:
                self.record_metric(metric, other.get_metric(metric))


class MetricsMixin:
    def __init__(self):
        super().__init__()
        self.compilation_metrics: Optional[CompilationMetrics] = None

    def are_metrics_enabled(self, metric_type: MetricsType = None):
        """
        Do we have a metrics collection, and if so does it have any active metrics.
        """
        return self.compilation_metrics is not None and (
            metric_type is None or self.compilation_metrics.are_enabled(metric_type)
        )

    def record_metric(self, metric: MetricsType, value):
        """
        Records a metric value if the collection has this sort of metric enabled.
        """
        if self.are_metrics_enabled(metric):
            self.compilation_metrics.record_metric(metric, value)

    def enable_metrics(self, enabled_metrics=None, overwrite=True):
        """
        Enables the set of metrics in the current collection. If overwrite is set to
        true, or there are no compilation metrics it'll create a new collection, if
        overwrite is false it'll enable these metrics in the currently-active
        collection.
        """
        if enabled_metrics is None:
            log.warning("Attempted to enable metrics with no value. Defaulting.")
            enabled_metrics = MetricsType.Default

        if self.compilation_metrics is None:
            return

        log.info(f"Enabling metrics with flags: {str(enabled_metrics)}")
        self.compilation_metrics.enable(enabled_metrics, overwrite)
